""":class:`GateGraph` is an alternative representation for PyRTL logic.

.. _gate_motivation:

Motivation
----------

.. doctest only::

    >>> import pyrtl
    >>> pyrtl.reset_working_block()

PyRTL represents logic internally with :class:`WireVectors<.WireVector>` and
:class:`LogicNets<.LogicNet>`. For example, the following code creates five
:class:`WireVectors<.WireVector>` and two :class:`LogicNets<.LogicNet>`::

    >>> a = pyrtl.Input(name="a", bitwidth=1)
    >>> b = pyrtl.Input(name="b", bitwidth=1)
    >>> c = pyrtl.Input(name="c", bitwidth=1)

    >>> x = a & b
    >>> x.name = "x"

    >>> y = x | c
    >>> y.name = "y"

    >>> print(pyrtl.working_block())
    x/1W <-- & -- a/1I, b/1I
    y/1W <-- | -- x/1W, c/1I

The :class:`WireVectors<.WireVector>` and :class:`LogicNets<.LogicNet>` are arranged
like this::

    ┌──────────────┐
    │ LogicNet "&" │
    │     op: "&"  │    ┌────────────────┐
    │     args:────┼───▶│ WireVector "a" │
    │     args:────┼─┐  └────────────────┘
    │              │ │  ┌────────────────┐
    │              │ └─▶│ WireVector "b" │
    │              │    └────────────────┘
    │              │    ┌────────────────┐
    │     dests:───┼───▶│ WireVector "x" │
    └──────────────┘ ┌─▶└────────────────┘
    ┌──────────────┐ │
    │ LogicNet "|" │ │
    │     op: "|"  │ │
    │     args:────┼─┘  ┌────────────────┐
    │     args:────┼───▶│ WireVector "c" │
    │              │    └────────────────┘
    │              │    ┌────────────────┐
    │     dests:───┼───▶│ WireVector "y" │
    └──────────────┘    └────────────────┘

This data structure is difficult to work with for three reasons:

1. The arrows do not consistently point from producer to consumer, or from consumer to
   producer. For example, there is no arrow from :class:`.WireVector` ``x`` (producer)
   to :class:`.LogicNet` ``|`` (consumer). Similarly, there is no arrow from
   :class:`.WireVector` ``x`` (consumer) to :class:`.LogicNet` ``&`` (producer). These
   missing arrows make it impossible to iteratively traverse the data structure. This
   creates a need for methods like :meth:`~.Block.net_connections`, which creates
   ``wire_src_dict`` and ``wire_sink_dict`` with the missing pointers.

2. The data structure is composed of two different classes, :class:`.LogicNet` and
   :class:`.WireVector`, and these two classes have completely different interfaces. As
   we follow pointers from one class to another, we must keep track of the current
   object's class, and interact with it appropriately.

3. :class:`.WireVector` is part of PyRTL's user interface, but also a key part of
   PyRTL's internal representation. This makes :class:`.WireVector` complex and
   difficult to modify, because it must implement user-facing features like inferring
   bitwidth from assignment, while also maintaining a consistent internal representation
   for simulation, analysis, and optimization.

:class:`GateGraph` is an alternative representation that addresses these issues. A
:class:`GateGraph` is just a collection of :class:`Gates<Gate>`, so we'll cover
:class:`Gate` first.
"""

from __future__ import annotations

from pyrtl.core import Block, LogicNet, working_block
from pyrtl.pyrtlexceptions import PyrtlError
from pyrtl.wire import Const, Input, Register, WireVector


class Gate:
    """:class:`Gate` is an alternative to PyRTL's default :class:`.LogicNet` and
    :class:`.WireVector` representation.

    :class:`Gate` makes it easy to iteratively explore a circuit, while simplifying the
    circuit's representation by making everything a :class:`Gate`. A :class:`Gate` is
    equivalent to a :class:`.LogicNet` fused with its :attr:`dest<.LogicNet.dests>`
    :class:`WireVector`. So this :class:`.LogicNet` and :class:`.WireVector`::

        ┌──────────────────┐
        │ LogicNet         │    ┌───────────────────┐
        │     op: o        │    │ WireVector        │
        │     args: [x, y] │    │     name: n       │
        │     dests:───────┼───▶│     bitwidth: b   │
        └──────────────────┘    └───────────────────┘

    Are equivalent to this :class:`Gate`::

        ┌─────────────────────┐
        │ Gate                │
        │     op: o           │
        │     args: [x, y]    │
        │     name: n         │
        │     bitwidth: b     │
        │     dests: [g1, g2] │
        └─────────────────────┘

    Key differences between the two representations:

    1. The :class:`Gate`'s :attr:`~Gate.args` ``[x, y]`` are references to other
       :class:`Gates<Gate>`.

    2. The :class:`.WireVector`'s :attr:`~WireVector.name` and
       :attr:`~WireVector.bitwidth` are stored as :attr:`Gate.name` and
       :attr:`Gate.bitwidth`. If the :class:`.LogicNet` produces no output, like a
       :class:`.MemBlock` write, the :class:`Gate`'s :attr:`~Gate.name` and
       :attr:`~Gate.bitwidth` will be ``None``. PyRTL does not have an
       :attr:`~.LogicNet.op` with multiple :attr:`.LogicNet.dests`.

    3. The :class:`Gate` has a new :attr:`Gate.dests` attribute, which has no direct
       equivalent in the :class:`.LogicNet`/:class:`.WireVector` representation.
       :attr:`Gate.dests` is a list of the :class:`Gates<Gate>` that use this
       :class:`Gate`'s output as one of their :attr:`~Gate.args`.

    :attr:`.LogicNet.dests` and :attr:`Gate.dests` represent slightly different things,
    despite having similar names:

    - :attr:`.LogicNet.dests` represents the :class:`LogicNet`'s output wire. It is a
      list of :class:`WireVectors<.WireVector>` which hold the :class:`.LogicNet`'s
      output. There can only be zero or one :class:`WireVectors<.WireVector>` in
      :attr:`.LogicNet.dests`, but that :class:`.WireVector` can be an
      :attr:`arg<.LogicNet.args>` to any number of :class:`LogicNets<.LogicNet>`.

    - :attr:`Gate.dests` represents the :class:`Gate`'s users. It is a list of
      :class:`Gates<Gate>` that use the :class:`Gate`'s output as one of their
      :attr:`~Gate.args`. There can be any number of :class:`Gates<Gate>` in
      :attr:`Gate.dests`.

    With :class:`Gates<Gate>`, the example from the :ref:`gate_motivation` section looks
    like::

        ┌─────────────────┐
        │ Gate "a"        │
        │     op: "I"     │
        │     name: "a"   │
        │     bitwidth: 1 │    ┌─────────────────┐
        │     dests:──────┼───▶│ Gate "&"        │
        └─────────────────┘◀─┐ │     op: "&"     │
        ┌─────────────────┐  └─┼─────args        │
        │ Gate "b"        │◀───┼─────args        │
        │     op: "I"     │    │     name: "x"   │    ┌─────────────────┐
        │     name: "b"   │    │     bitwidth: 1 │    │ Gate "|"        │
        │     bitwidth: 1 │ ┌─▶│     dests:──────┼───▶│     op: "|"     │
        │     dests:──────┼─┘  └─────────────────┘◀───┼─────args        │
        └─────────────────┘  ┌────────────────────────┼─────args        │
        ┌─────────────────┐  │                        │     name: "y"   │
        │ Gate "c"        │◀─┘   ┌───────────────────▶│     bitwidth: 1 │
        │     op: "I"     │      │                    └─────────────────┘
        │     name: "c"   │      │
        │     bitwidth: 1 │      │
        │     dests:──────┼──────┘
        └─────────────────┘

    With a :class:`Gate` representation, it is easy to iteratively traverse the data
    structure:

    1. Forwards, from producer to consumer, by following :attr:`~Gate.dests` references.

    2. Backwards, from consumer to producer, by following :attr:`~Gate.args` references.

    The :class:`Gate` representation addresses the issues raised in the
    :ref:`gate_motivation` section:

    1. The :class:`Gate` representation is easy to iteratively explore by following
       :attr:`~Gate.args` and :attr:`~Gate.dests` references, which are shown as arrows
       in the figure above.

    2. There is only one class in the :class:`Gate` graph, so we don't need to keep
       track of the current object's type as we follow arrows in the graph, like we did
       with :class:`LogicNet` and :class:`WireVector`. Everything is a :class:`Gate`.

    3. By decoupling the :class:`Gate` representation from :class:`.WireVector` and
       :class:`.LogicNet`, :class:`Gate` specializes in supporting analysis use cases,
       without the burden of supporting all of :class:`.WireVector`'s other features.
       This significantly simplifies :class:`Gate`'s design and implementation.

    For usage examples, see :class:`GateGraph` and :class:`Gate`'s documentation below.
    """

    op: str
    """Operation performed by this ``Gate``. Corresponds to :attr:`.LogicNet.op`.

    For special ``Gates`` created for :class:`Inputs<.Input>`, ``op`` will instead be
    the :class:`.Input`'s ``_code``, which is ``I``.

    For special ``Gates`` created for :class:`Consts<.Const>`, ``op`` will instead be
    the :class:`.Const`'s ``_code``, which is ``C``.

    See :class:`.LogicNet`'s documentation for a description of all other ``ops``.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> a = pyrtl.Input(name="a", bitwidth=1)
        >>> _ = ~a

        >>> gate_graph = pyrtl.GateGraph()
        >>> gate_a = gate_graph.get_gate("a")
        >>> gate_a.op
        'I'
        >>> gate_a.dests[0].op
        '~'
    """

    op_param: tuple
    """Static parameters for the operation. Corresponds to :attr:`.LogicNet.op_param`.

    These are constant parameters, whose values are statically known. These values
    generally do not appear as actual values on wires. For example, the bits to select
    for the ``s`` bit-slice operation are stored as ``op_params``.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> a = pyrtl.Input(name="a", bitwidth=8)
        >>> bit_slice = a[1:3]
        >>> bit_slice.name = "bit_slice"

        >>> gate_graph = pyrtl.GateGraph()
        >>> bit_slice_gate = gate_graph.get_gate("bit_slice")
        >>> bit_slice_gate.op_param
        (1, 2)
    """

    args: list[Gate]
    """Inputs to the operation. Corresponds to :attr:`.LogicNet.args`.

    For each ``Gate`` ``arg`` in ``self.args``, ``self`` is in ``arg.dests``.

    Some special ``Gates`` represent operations without ``args``, like :class:`.Input`
    and :class:`.Const`. Such operations will have an empty list of ``args``.

    .. note::

        The same ``Gate`` may appear multiple times in ``args``. A self-loop
        :class:`.Register` ``Gate`` may be its own ``arg``.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> a = pyrtl.Input(name="a", bitwidth=1)
        >>> b = pyrtl.Input(name="b", bitwidth=1)
        >>> c = pyrtl.Input(name="c", bitwidth=1)
        >>> abc = pyrtl.concat(a, b, c)
        >>> abc.name = "abc"

        >>> gate_graph = pyrtl.GateGraph()
        >>> abc_gate = gate_graph.get_gate("abc")
        >>> [gate.name for gate in abc_gate.args]
        ['a', 'b', 'c']
    """

    name: str | None
    """Name of the operation's output. Corresponds to :attr:`.WireVector.name`.

    Some operations do not have outputs, like :class:`.MemBlock` writes. These
    operations will have a ``name`` of ``None``.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> a = pyrtl.Input(name="a", bitwidth=1)
        >>> b = pyrtl.Input(name="b", bitwidth=1)
        >>> ab = a + b
        >>> ab.name = "ab"

        >>> gate_graph = pyrtl.GateGraph()
        >>> ab_gate = gate_graph.get_gate("ab")
        >>> ab_gate.name
        'ab'
    """

    bitwidth: int | None
    """Bitwidth of the operation's output. Corresponds to :attr:`.WireVector.bitwidth`.

    Some operations do not have outputs, like :class:`.MemBlock` writes. These
    operations will have a ``bitwidth`` of ``None``.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> a = pyrtl.Input(name="a", bitwidth=1)
        >>> b = pyrtl.Input(name="b", bitwidth=1)
        >>> ab = a + b
        >>> ab.name = "ab"

        >>> gate_graph = pyrtl.GateGraph()
        >>> ab_gate = gate_graph.get_gate("ab")
        >>> ab_gate.bitwidth
        2
    """

    dests: list[Gate]
    """:class:`list` of :class:`Gates<Gate>` that use this operation's output as one of
    their :attr:`~Gate.args`.

    For each :class:`Gate` ``dest`` in ``self.dests``, ``self`` is in ``dest.args``.

    .. note::

        The same :class:`Gate` may appear multiple times in ``dests``. A self-loop
        :class:`.Register` ``Gate`` may appear in its own ``dests``.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> a = pyrtl.Input(name="a", bitwidth=1)
        >>> _ = a + 1
        >>> _ = a - 1

        >>> gate_graph = pyrtl.GateGraph()
        >>> a_gate = gate_graph.get_gate("a")
        >>> [gate.op for gate in a_gate.dests]
        ['+', '-']
    """

    is_output: bool
    """Indicates if the operation's output is an :class:`.Output`.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> a = pyrtl.Input(name="a", bitwidth=1)
        >>> b = pyrtl.Output(name="b", bitwidth=1)
        >>> b <<= a

        >>> gate_graph = pyrtl.GateGraph()
        >>> a_gate = gate_graph.get_gate("a")
        >>> a_gate.is_output
        False
        >>> b_gate = gate_graph.get_gate("b")
        >>> b_gate.is_output
        True
    """

    def __init__(
        self,
        logic_net: LogicNet = None,
        wire_vector: WireVector = None,
        args: list[Gate] | None = None,
    ):
        """Create a ``Gate`` from a :class:`.LogicNet` or :class:`.WireVector`.

        ``Gates`` are complicated to construct because they are doubly-linked, and there
        may be cycles in the ``Gate`` graph. Most users should not call this constructor
        directly, and instead use :class:`GateGraph` to create ``Gates`` from a
        :class:`.Block`.

        :param logic_net: :class:`.LogicNet` to create this ``Gate`` from. If
            ``logic_net`` is specified, ``wire_vector`` must be ``None``.

            ``logic_net`` must not be a register, where ``logic_net.op == 'r'``.

            Register ``Gates`` are created in two phases by :class:`GateGraph`. In the
            first phase, a placeholder ``Gate`` is created from the :class:`.Register`.
            In this first phase, the register ``Gate``'s ``op`` is temporarily set to
            ``R``, which is the :class:`.Register`'s ``_code``. This placeholder is
            needed to resolve other ``Gate``'s references to the register in the second
            phase. In the second phase, the register ``Gate``'s remaining fields are
            populated from the register's :class:`.LogicNet`. In the second phase, the
            register ``Gate``'s ``op`` is changed to ``r``, which is the
            :class:`.LogicNet`'s :attr:`~.LogicNet.op`.

        :param wire_vector: :class:`.WireVector` to create this ``Gate`` from. If
            ``wire_vector`` is specified, ``logic_net`` must be ``None``.

            ``wire_vector`` must be a :class:`.Const`, :class:`.Input`, or
            :class:`.Register`.

        :param args: A :class:`list` of ``Gates`` that are inputs to this ``Gate``. This
            corresponds to :attr:`.LogicNet.args`, except that each of a ``Gate``'s
            ``args`` is a ``Gate``.
        """
        self.op_param = None
        if args is None:
            self.args = []
        else:
            self.args = args
        self.name = None
        self.bitwidth = None
        # ``dests`` will be set up later, by ``GateGraph``.
        self.dests = []
        self.is_output = False

        if logic_net is not None:
            # Constructing a ``Gate`` from a ``logic_net``.
            #
            # For ``LogicNets``, set the ``Gate``'s ``op``, ``op_param``, ``name``,
            # ``bitwidth``.
            if wire_vector is not None:
                msg = "Do not pass both logic_net and wire_vector to Gate."
                raise PyrtlError(msg)
            self.op = logic_net.op
            if self.op == "r":
                msg = "Registers should be created from a wire_vector, not a logic_net."
                raise PyrtlError(msg)
            self.op_param = logic_net.op_param

            num_dests = len(logic_net.dests)
            if num_dests:
                if num_dests > 1:
                    # The ``Gate`` representation supports at most one ``LogicNet``
                    # ``dest``. If more than one ``LogicNet`` ``dest`` is needed in the
                    # future, concat them together, then split them apart with ``s``
                    # bit-selection ``Gates``, or use multiple ``Gates`` with the same
                    # ``args``.
                    msg = "LogicNets with more than one dest are not supported"
                    raise PyrtlError(msg)
                dest = logic_net.dests[0]
                self.name = dest.name
                self.bitwidth = dest.bitwidth
                if dest._code == "O":
                    self.is_output = True

        else:
            # Constructing a ``Gate`` from a ``wire_vector``.
            #
            # For ``Inputs`` and ``Registers``, set the ``Gate``'s ``op`` and ``dest``.
            # For ``Consts``, also copy the ``val`` to ``op_param``.
            # For ``Registers``, also copy the ``reset_value`` to ``op_param``.
            if wire_vector is None:
                msg = "Gate must be constructed from a logic_net or a wire_vector."
                raise PyrtlError(msg)

            if wire_vector._code not in "CIR":
                msg = (
                    "Gate must be constructed from a Const, Input or Register "
                    "wire_vector."
                )
                raise PyrtlError(msg)

            self.op = wire_vector._code
            self.name = wire_vector.name
            self.bitwidth = wire_vector.bitwidth
            if self.op == "C":
                self.op_param = (wire_vector.val,)
            elif self.op == "R":
                if not wire_vector.reset_value:
                    self.op_param = (0,)
                else:
                    self.op_param = (wire_vector.reset_value,)

    def __str__(self) -> str:
        """:return: A string representation of this ``Gate``.

        .. doctest only::

            >>> import pyrtl
            >>> pyrtl.reset_working_block()

        Example::

            >>> a = pyrtl.Input(name="a", bitwidth=8)
            >>> bit_slice = a[2:4]
            >>> bit_slice.name = "bit_slice"

            >>> gate_graph = pyrtl.GateGraph()
            >>> bit_slice_gate = gate_graph.get_gate("bit_slice")

            >>> print(bit_slice_gate)
            bit_slice/2 = slice(a/8) [sel=(2, 3)]

        In this sample string representation:

        - :attr:`~Gate.name` is ``bit_slice``.

        - :attr:`~Gate.bitwidth` is ``2``.

        - :attr:`~Gate.op` is ``s``, spelled out as ``slice`` to improve readability.

        - :attr:`~Gate.args` is ``[<Gate for "a">]``::

            >>> bit_slice_gate.args[0] is gate_graph.get_gate("a")
            True

        - :attr:`~Gate.op_param` is ``(2, 3)``, written as ``sel``
          because a ``slice``'s :attr:`~Gate.op_param` determines the selected bits.
          This improves readability by indicating what the :attr:`~Gate.op_param` means
          for the :attr:`~Gate.op`.
        """
        if self.name is None:
            dest = ""
        else:
            dest_notes = ""
            if self.is_output:
                dest_notes = " [Output]"

            dest = f"{self.name}/{self.bitwidth}{dest_notes} "

        op_name_map = {
            "&": "and",
            "|": "or",
            "^": "xor",
            "n": "nand",
            "~": "invert",
            "+": "add",
            "-": "sub",
            "*": "mul",
            "=": "eq",
            "<": "lt",
            ">": "gt",
            "w": "",
            "x": "",  # Multiplexers are printed as ternary operators.
            "c": "concat",
            "s": "slice",
            "r": "reg",
            "m": "read",
            "@": "write",
            "I": "Input",
            "C": "Const",
        }
        if self.name is None:
            op = op_name_map[self.op]
        else:
            op = f"= {op_name_map[self.op]}"

        if not self.args:
            args = ""
        else:
            arg_names = [f"{arg.name}/{arg.bitwidth}" for arg in self.args]
            if self.op == "w":
                args = arg_names[0]
            elif self.op == "x":
                args = f"{arg_names[0]} ? {arg_names[2]} : {arg_names[1]}"
            elif self.op == "m":
                args = f"(addr={arg_names[0]})"
            elif self.op == "@":
                args = (
                    f"(addr={arg_names[0]}, data={arg_names[1]}, enable={arg_names[2]})"
                )
            else:
                args = f"({', '.join(arg_names)})"

        if self.op_param is None:
            op_param = ""
        elif self.op == "C":
            op_param = f"({self.op_param[0]})"
        elif self.op == "s":
            op_param = f" [sel={self.op_param}]"
        elif self.op == "m" or self.op == "@":
            op_param = f" [memid={self.op_param[0]} mem={self.op_param[1].name}]"
        elif self.op == "r":
            op_param = f" [reset_value={self.op_param[0]}]"
        else:
            op_param = f" [op_param={self.op_param}]"

        return f"{dest}{op}{args}{op_param}"


class GateGraph:
    """A :class:`GateGraph` is a collection of :class:`Gates<Gate>`.
    :class:`GateGraph`'s constructor creates :class:`Gates<Gate>` from a
    :class:`.Block`.

    See :ref:`gate_motivation` for more background.

    Users should generally construct :class:`GateGraphs<GateGraph>`, rather than
    attempting to directly construct individual :class:`Gates<Gate>`. :class:`Gate`
    construction is complex because they are doubly-linked, and the :class:`Gate` graph
    may contain cycles.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example
    -------

    Let's build a :class:`GateGraph` for the :ref:`gate_motivation` example::

        >>> a = pyrtl.Input(name="a", bitwidth=1)
        >>> b = pyrtl.Input(name="b", bitwidth=1)
        >>> c = pyrtl.Input(name="c", bitwidth=1)

        >>> x = a & b
        >>> x.name = "x"

        >>> y = x | c
        >>> y.name = "y"

        >>> gate_graph = pyrtl.GateGraph()

    The :class:`GateGraph` can be printed, revealing five :class:`Gates<Gate>`::

        >>> print(gate_graph)
        a/1 = Input
        b/1 = Input
        c/1 = Input
        x/1 = and(a/1, b/1)
        y/1 = or(x/1, c/1)

    We can retrieve the :attr:`Gate` for input ``a``::

        >>> a = gate_graph.get_gate("a")
        >>> print(a)
        a/1 = Input
        >>> a.name
        'a'
        >>> a.op
        'I'

    We can check ``a``'s :attr:`~Gate.dests` to see that it is an argument to a bitwise
    ``&`` operation, with output named ``x``::

        >>> len(a.dests)
        1
        >>> x = a.dests[0]
        >>> print(x)
        x/1 = and(a/1, b/1)
        >>> x.op
        '&'
        >>> x.name
        'x'

    We can examine the bitwise ``&``'s :attr:`~Gate.args`, to get references to input
    :class:`Gates<Gate>` ``a`` and ``b``::

        >>> x.args[0] is a
        True

        >>> b = x.args[1]
        >>> print(b)
        b/1 = Input
        >>> b.name
        'b'
        >>> b.op
        'I'

    Special :class:`Gates<Gate>`
    ----------------------------

    Generally, :class:`GateGraph` converts each :class:`.LogicNet` in a :class:`.Block`
    to a corresponding :class:`Gate`, but some :class:`WireVectors<.WireVector>` and
    :class:`LogicNets<.LogicNet>` are handled differently:

    - An :class:`.Input` :class:`.WireVector` is converted to a special input
      :class:`Gate`, with op ``I``. Input :class:`Gates<Gate>` have no
      :attr:`~Gate.args`, and do not correspond to a :class:`.LogicNet`.

    - A :class:`.Const` :class:`.WireVector` is converted to a special const
      :class:`Gate`, with op ``C``. Const :class:`Gates<Gate>` have no
      :attr:`~Gate.args`, and do not correspond to a :class:`.LogicNet`. The constant's
      value is stored in :attr:`Gate.op_param`.

    - An :class:`.Output` :class:`.WireVector` is handled normally, and will be the
      ``dest`` of the :class:`Gate` that defines the :class:`.Output`'s value. That
      :class:`Gate` will have its :attr:`~Gate.is_output` attribute set to ``True``.

    - :class:`.Register` :class:`WireVectors<.WireVector>` and
      :class:`LogicNets<.LogicNet>` are handled normally, except that the
      :class:`.Register`'s ``reset_value`` is stored in :attr:`Gate.op_param`. Register
      :class:`Gates<Gate>` use the register :class:`.LogicNet` :attr:`~.LogicNet.op`
      ``r``, not the :class:`.Register` ``_code`` ``R``.

    .. note::

        Registers can create cycles in the :class:`Gate` graph, because the logic that
        defines the register's :attr:`~.Register.next` value (which is the register
        :class:`Gate`'s :attr:`~Gate.args`) can depend on the register's current value
        (which is the register :class:`Gate`'s :attr:`~Gate.dests`). Watch out for
        infinite loops when traversing a :class:`GateGraph` with registers. For example,
        if you keep following :attr:`~Gate.dests` references, you may end up back where
        you started.
    """

    gates: set[Gate]
    """A :class:`set` of all :class:`Gates<Gate>` in the ``GateGraph``.

    Similar to :attr:`~GateGraph.__iter__`, except that ``gates`` is a :class:`set`
    rather than an :class:`~collections.abc.Iterable`.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> a = pyrtl.Input(name="a", bitwidth=1)
        >>> b = pyrtl.Input(name="b", bitwidth=1)
        >>> c = pyrtl.Input(name="c", bitwidth=1)
        >>> x = a & b
        >>> x.name = "x"
        >>> y = x | c
        >>> y.name = "y"

        >>> gate_graph = pyrtl.GateGraph()

        >>> sorted(gate.name for gate in gate_graph.gates)
        ['a', 'b', 'c', 'x', 'y']
    """

    consts: set[Gate]
    """A :class:`set` of :class:`.Const` :class:`Gates<Gate>` in the ``GateGraph``.

    These :class:`Gates<Gate>` provide constant values, with :attr:`~Gate.op` ``C``.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> c = pyrtl.Const(name="c", val=0)
        >>> d = pyrtl.Const(name="d", val=1)
        >>> _ = c + d

        >>> gate_graph = pyrtl.GateGraph()

        >>> sorted(gate.name for gate in gate_graph.consts)
        ['c', 'd']
    """

    inputs: set[Gate]
    """A :class:`set` of :class:`.Input` :class:`Gates<Gate>` in the ``GateGraph``.

    These :class:`Gates<Gate>` provide :class:`.Input` values, with :attr:`~Gate.op`
    ``I``.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> a = pyrtl.Input(name="a", bitwidth=1)
        >>> b = pyrtl.Input(name="b", bitwidth=1)
        >>> _ = a & b

        >>> gate_graph = pyrtl.GateGraph()

        >>> sorted(gate.name for gate in gate_graph.inputs)
        ['a', 'b']
    """

    outputs: set[Gate]
    """A :class:`set` of :class:`.Output` :class:`Gates<Gate>` in the ``GateGraph``.

    These :class:`Gates<Gate>` set :class:`.Output` values, with :attr:`~Gate.is_output`
    ``True``.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> x = pyrtl.Output(name="x")
        >>> y = pyrtl.Output(name="y")
        >>> x <<= 42
        >>> y <<= 255

        >>> gate_graph = pyrtl.GateGraph()

        >>> sorted(gate.name for gate in gate_graph.outputs)
        ['x', 'y']
    """

    registers: set[Gate]
    """A :class:`set` of :class:`.Register` update :class:`Gates<Gate>` in the
    ``GateGraph``.

    These :class:`Gates<Gate>` set a :class:`.Register`'s value for the next cycle, with
    :attr:`~Gate.op` ``r``.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> r = pyrtl.Register(name="r", bitwidth=1)
        >>> s = pyrtl.Register(name="s", bitwidth=1)
        >>> r.next <<= r + 1
        >>> s.next <<= s + 2

        >>> gate_graph = pyrtl.GateGraph()

        >>> sorted(gate.name for gate in gate_graph.registers)
        ['r', 's']
    """

    mem_reads: set[Gate]
    """A :class:`set` of :class:`.MemBlock` read :class:`Gates<Gate>` in the
    ``GateGraph``.

    These :class:`Gates<Gate>` read :class:`MemBlocks<.MemBlock>`, with
    :attr:`~Gate.op` ``m``.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> mem = pyrtl.MemBlock(name="mem", bitwidth=4, addrwidth=2)
        >>> addr = pyrtl.Input(name="addr", bitwidth=2)
        >>> mem_read_1 = mem[addr]
        >>> mem_read_1.name = "mem_read_1"
        >>> mem_read_2 = mem[addr]
        >>> mem_read_2.name = "mem_read_2"

        >>> gate_graph = pyrtl.GateGraph()

        >>> sorted(gate.name for gate in gate_graph.reads)
        ['mem_read_1', 'mem_read_2']
    """

    mem_writes: set[Gate]
    """A :class:`set` of :class:`.MemBlock` write :class:`Gates<Gate>` in the
    ``GateGraph``.

    These :class:`Gates<Gate>` write :class:`MemBlocks<.MemBlock>`, with
    :attr:`~Gate.op` ``@``.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> mem = pyrtl.MemBlock(name="mem", bitwidth=4, addrwidth=2)
        >>> addr = pyrtl.Input(name="addr", bitwidth=2)
        >>> mem[addr] <<= 7

        >>> gate_graph = pyrtl.GateGraph()

        >>> # MemBlock writes have no name.
        >>> [gate.name for gate in gate_graph.mem_writes]
        [None]

        >>> [gate.op for gate in gate_graph.mem_writes]
        ['@']
    """

    sources: set[Gate]
    """A :class:`set` of ``source`` :class:`Gates<Gate>` in the ``GateGraph``.

    A ``source`` :class:`Gate`'s output value is known at the beginning of each clock
    cycle. :class:`Consts<.Const>`, :class:`Inputs<.Input>`, and
    :class:`Registers<.Register>` are ``source`` :class:`Gates<Gate>`.

    .. note::

        :class:`Registers<.Register>` are both ``sources`` and :attr:`~GateGraph.sinks`.
        As a ``source``, it provides the :class:`.Register`'s value for the current
        cycle. As a :attr:`sink<GateGraph.sinks>`, it determines the
        :class:`.Register`'s value for the next cycle.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> a = pyrtl.Input(name="a", bitwidth=1)
        >>> c = pyrtl.Const(name="c", bitwidth=1, val=0)
        >>> r = pyrtl.Register(name="r", bitwidth=1)
        >>> r.next <<= a + c

        >>> gate_graph = pyrtl.GateGraph()

        >>> sorted(gate.name for gate in gate_graph.sources)
        ['a', 'c', 'r']
    """

    sinks: set[Gate]
    """A :class:`set` of ``sink`` :class:`Gates<Gate>` in the ``GateGraph``.

    A ``sink`` :class:`Gate`'s output value is known only at the end of each clock
    cycle. :class:`Registers<.Register>`, :class:`Outputs<.Output>`, :class:`MemBlock`
    writes, and any :class:`Gate` without users (``len(dests) == 0``) are sink
    :class:`Gates<Gate>`.

    .. note::

        :class:`Registers<.Register>` are both :attr:`~GateGraph.sources` and ``sinks``.
        As a :attr:`source<GateGraph.sources>`, it provides the :class:`.Register`'s
        value for the current cycle. As a ``sink``, it determines the
        :class:`.Register`'s value for the next cycle.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> a = pyrtl.Input(name="a", bitwidth=1)
        >>> r = pyrtl.Register(name="r", bitwidth=1)
        >>> o = pyrtl.Output(name="o", bitwidth=1)
        >>> r.next <<= a + 1
        >>> o <<= 1
        >>> sum = a + r
        >>> sum.name = "sum"

        >>> gate_graph = pyrtl.GateGraph()

        >>> sorted(gate.name for gate in gate_graph.sinks)
        ['o', 'r', 'sum']
    """

    def __init__(self, block: Block = None):
        """Create :class:`Gates<Gate>` from a :class:`.Block`.

        Most users should call this constructor, rather than attempting to directly
        construct individual :class:`Gates<Gate>`.

        :param block: :class:`.Block` to construct the :class:`GateGraph` from. Defaults
            to the :ref:`working_block`.
        """
        self.gates = set()
        self.consts = set()
        self.inputs = set()
        self.outputs = set()
        self.registers = set()
        self.mem_reads = set()
        self.mem_writes = set()
        self.sources = set()
        self.sinks = set()

        block = working_block(block)
        block.sanity_check()

        # The ``Gate`` graph is doubly-linked, and may contain cycles, so construction
        # is done in two phases. In the first phase, we only construct ``Gates`` for
        # ``sources``, which are ``Consts``, ``Inputs``, and ``Registers``.
        #
        # In this phase, register gates are placeholders. They do not have ``args``, and
        # their ``op`` is temporarily ``R``, which is ``Register._code`. These
        # placeholders are needed to resolve references to registers in the second
        # phase.
        #
        # ``wire_vector_map`` maps from ``WireVector`` to the corresponding gate. It is
        # initially populated with ``Gates`` constructed from ``sources``.
        wire_vector_map: dict[WireVector, Gate] = {}
        for wire_vector in block.wirevector_subset((Const, Input, Register)):
            gate = Gate(wire_vector=wire_vector)
            self.gates.add(gate)
            self.sources.add(gate)
            wire_vector_map[wire_vector] = gate

            if gate.op == "C":
                self.consts.add(gate)
            elif gate.op == "I":
                self.inputs.add(gate)
            elif gate.op == "R":
                self.registers.add(gate)

        # In the second phase, we construct all remaining ``Gates`` from ``LogicNets``.
        # ``Block``'s iterator returns ``LogicNets`` in topological order, so we can be
        # sure that each ``LogicNet``'s ``args`` are all in ``wire_vector_map``.
        for logic_net in block:
            # Find the ``Gates`` corresponding to the ``LogicNet``'s ``args``.
            gate_args = []
            for wire_arg in logic_net.args:
                gate_arg = wire_vector_map.get(wire_arg)
                if gate_arg is None:
                    msg = f"Missing Gate for wire {wire_arg}"
                    raise PyrtlError(msg)
                gate_args.append(gate_arg)
            if logic_net.op == "r":
                # Find the placeholder register ``Gate`` we created earlier, and finish
                # constructing it.
                gate = wire_vector_map[logic_net.dests[0]]
                gate.op = "r"
                gate.args = gate_args
                self.sinks.add(gate)
            else:
                gate = Gate(logic_net=logic_net, args=gate_args)
                self.gates.add(gate)

            # Add the new ``Gate`` as a ``dest`` for its ``args``.
            for gate_arg in gate_args:
                gate_arg.dests.append(gate)

            # Add the new ``Gate`` to ``wire_vector_map``, so we can resolve future
            # references to it.
            num_dests = len(logic_net.dests)
            if num_dests:
                if num_dests > 1:
                    msg = "LogicNets with more than one dest are not supported"
                    raise PyrtlError(msg)
                dest = logic_net.dests[0]
                wire_vector_map[dest] = gate

            if gate.is_output:
                self.outputs.add(gate)
            if gate.op == "m":
                self.mem_reads.add(gate)
            elif gate.op == "@":
                self.mem_writes.add(gate)

        for gate in self.gates:
            if len(gate.dests) == 0:
                self.sinks.add(gate)

    def get_gate(self, name: str) -> Gate | None:
        """Return the :class:`Gate` whose :attr:`~Gate.name` is ``name``, or ``None`` if
        no such :class:`Gate` exists.

        .. warning::

            :class:`.MemBlock` writes do not produce an output, so they can not be
            retrieved with ``get_gate``.

        .. doctest only::

            >>> import pyrtl
            >>> pyrtl.reset_working_block()

        Example::

            >>> a = pyrtl.Input(name="a", bitwidth=1)
            >>> na = ~a
            >>> na.name = "na"

            >>> gate_graph = pyrtl.GateGraph()

            >>> a_gate = gate_graph.get_gate("a")
            >>> na_gate = gate_graph.get_gate("na")
            >>> na_gate.op
            '~'
            >>> na_gate.args[0] is a_gate
            True

        :param name: Name of the :class:`Gate` to find.

        :return: The named :class:`Gate`, or ``None`` if no such :class:`Gate` was
                 found.
        """
        for gate in self.gates:
            if gate.name == name:
                return gate
        return None

    def __str__(self) -> str:
        """Return a string representation of the ``GateGraph``.

        .. doctest only::

            >>> import pyrtl
            >>> pyrtl.reset_working_block()

        Example::

            >>> a = pyrtl.Input(name="a", bitwidth=2)
            >>> b = pyrtl.Input(name="b", bitwidth=2)
            >>> sum = a + b
            >>> sum.name = "sum"

            >>> gate_graph = pyrtl.GateGraph()

            >>> print(gate_graph)
            a/2 = Input
            b/2 = Input
            sum/3 = add(a/2, b/2)

        :return: A string representation of each :class:`Gate` in the ``GateGraph``, one
                 :class:`Gate` per line. The :class:`Gates<Gate>` will be sorted by
                 name.
        """
        sorted_gates = sorted(
            self.gates, key=lambda gate: gate.name if gate.name else "~~~"
        )
        return "\n".join([str(gate) for gate in sorted_gates])

    def __iter__(self):
        """Iterate over each gate in the :class:`GateGraph`.

        Similar to :attr:`~GateGraph.gates`, except that ``__iter__`` returns an
        :class:`~collections.abc.Iterable` rather than a :class:`set`.

        .. doctest only::

            >>> import pyrtl
            >>> pyrtl.reset_working_block()

        Example::

            >>> a = pyrtl.Input(name="a", bitwidth=2)
            >>> b = pyrtl.Input(name="b", bitwidth=2)
            >>> sum = a + b
            >>> sum.name = "sum"

            >>> gate_graph = pyrtl.GateGraph()

            >>> sorted(gate.name for gate in gate_graph)
            ['a', 'b', 'sum']
        """
        return iter(self.gates)



"""Helper functions that make constructing hardware easier."""

from __future__ import annotations

import collections
import numbers
import random
from typing import TYPE_CHECKING, NamedTuple

from pyrtl.core import Block, _get_debug_mode, _NameIndexer, working_block
from pyrtl.corecircuits import (
    as_wires,
    concat,
    concat_list,
    rtl_all,
    rtl_any,
    select,
    shift_left_logical,
)
from pyrtl.pyrtlexceptions import PyrtlError, PyrtlInternalError
from pyrtl.wire import (
    Const,
    Input,
    Output,
    Register,
    WireVector,
    WireVectorLike,
    WrappedWireVector,
)

if TYPE_CHECKING:
    from pyrtl.simulation import Simulation


# -----------------------------------------------------------------
#        ___       __   ___  __   __
#  |__| |__  |    |__) |__  |__) /__`
#  |  | |___ |___ |    |___ |  \ .__/
#
probeIndexer = _NameIndexer("Probe-")


def probe(w: WireVector, name: str | None = None) -> WireVector:
    """Print useful information about a :class:`WireVector` in debug mode.

    ``probe`` can be inserted into a existing design easily because it returns the
    original wire unmodified. For example::

        y <<= x[0:3] + 4

    could be rewritten as::

        y <<= probe(x)[0:3] + 4

    to give visibility into both the origin of ``x`` (including the line that
    :class:`WireVector` was originally created) and the run-time values of ``x`` (which
    will be named and thus show up by default in a trace). Likewise::

        y <<= probe(x[0:3]) + 4
        y <<= probe(x[0:3] + 4)
        probe(y) <<= x[0:3] + 4

    are all valid uses of ``probe``.

    .. note::

        ``probe`` actually adds an :class:`Output` wire to the :ref:`working_block` of
        ``w``, which can confuse various post-processing transforms such as
        :func:`output_to_verilog`.

    :param w: :class:`WireVector` from which to get info
    :param name: optional name for probe (defaults to an autogenerated name)

    :return: original :class:`WireVector` ``w``
    """
    if not isinstance(w, WireVector):
        msg = "Only WireVectors can be probed"
        raise PyrtlError(msg)

    if name is None:
        name = f"({probeIndexer.make_valid_string()}: {w.name})"
    if _get_debug_mode():
        print("Probe: " + name + " " + get_stack(w))

    p = Output(name=name)
    p <<= w  # late assigns len from w automatically
    return w


assertIndexer = _NameIndexer("assertion")


def rtl_assert(w: WireVector, exp: Exception, block: Block = None) -> Output:
    """Add a hardware assertion to be checked during simulation.

    If at any time during execution the wire ``w`` is not ``True`` (i.e. when it is
    asserted low) then :class:`Simulation` will raise ``exp``.

    :param w: A 1-bit :class:`WireVector` to assert.
    :param exp: :class:`Exception` to throw when the assertion fails during
        :class:`Simulation`.
    :param block: :class:`Block` to which the assertion should be added (default to
        :ref:`working_block`).

    :raises exp: When ``w`` is not ``True``.

    :return: The :class:`Output` wire for the assertion, which can be ignored in most
             cases.
    """
    block = working_block(block)

    if not isinstance(w, WireVector):
        msg = "Only WireVectors can be asserted with rtl_assert"
        raise PyrtlError(msg)
    if len(w) != 1:
        msg = "rtl_assert checks only a WireVector of bitwidth 1"
        raise PyrtlError(msg)
    if not isinstance(exp, Exception):
        msg = "the second argument to rtl_assert must be an instance of Exception"
        raise PyrtlError(msg)
    if isinstance(exp, KeyError):
        msg = "the second argument to rtl_assert cannot be a KeyError"
        raise PyrtlError(msg)
    if w not in block.wirevector_set:
        msg = "assertion wire not part of the block to which it is being added"
        raise PyrtlError(msg)
    if w not in block.wirevector_set:
        msg = "assertion not a known wirevector in the target block"
        raise PyrtlError(msg)

    if w in block.rtl_assert_dict:
        msg = "assertion conflicts with existing registered assertion"
        raise PyrtlInternalError(msg)

    assert_wire = Output(
        bitwidth=1, name=assertIndexer.make_valid_string(), block=block
    )
    assert_wire <<= w
    block.rtl_assert_dict[assert_wire] = exp
    return assert_wire


def check_rtl_assertions(sim: Simulation):
    """Checks for failing assertions in ``sim``.

    :class:`Simulation` calls this automatically, so users generally shouldn't need to
    call this. See :func:`rtl_assert`.

    :param sim: Simulation in which to check assertions.
    """
    for w, exp in sim.block.rtl_assert_dict.items():
        try:
            value = sim.inspect(w)
            if not value:
                raise exp
        except KeyError:
            pass


def log2(integer_val: int) -> int:
    """Return the base-2 logarithm of an integer.

    Useful when checking that powers of 2 are provided as function inputs.

    Examples::

        >>> log2(2)
        1
        >>> log2(256)
        8
        >>> log2(100)
        Traceback (most recent call last):
          ...
        pyrtl.pyrtlexceptions.PyrtlError: this function can only take even powers of 2

    :param integer_val: The integer to take the log base 2 of.

    :raise PyrtlError: If the input is negative, or not an even power of 2.

    :return: The log base 2 of ``integer_val``.
    """
    i = integer_val
    if not isinstance(i, int):
        msg = "this function can only take integers"
        raise PyrtlError(msg)
    if i <= 0:
        msg = "this function can only take positive numbers 1 or greater"
        raise PyrtlError(msg)
    if i & (i - 1) != 0:
        msg = "this function can only take even powers of 2"
        raise PyrtlError(msg)
    return i.bit_length() - 1


def truncate(
    wirevector_or_integer: WireVector | int, bitwidth: int
) -> WireVector | int:
    """Returns a :class:`WireVector` or integer truncated to the specified ``bitwidth``.

    Truncation removes the most significant bits of ``wirevector_or_integer``, leaving a
    result that is only :attr:`~WireVector.bitwidth` bits wide. For :class:`ints<int>`
    this is performed with a simple bitmask of size :attr:`~WireVector.bitwidth`, and
    returns an :class:`int`. For :class:`WireVectors<WireVector>` the function calls
    :meth:`WireVector.truncate` and returns a :class:`WireVector` with the specified
    :attr:`~WireVector.bitwidth`.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Examples::

        >>> truncate(0b101_001, bitwidth=3)
        1

        >>> bin(truncate(0b111_101, bitwidth=3))
        '0b101'

        >>> # -1 is 0b1111111... with the number of 1-bits equal to the bitwidth. Python
        >>> # ints are arbitrary-precision, so this can produce any number of 1-bits.
        >>> bin(truncate(-1, bitwidth=3))
        '0b111'

        >>> input = pyrtl.Input(name="input", bitwidth=8)

        >>> output = truncate(input, bitwidth=4)
        >>> output.name = "output"
        >>> output.bitwidth
        4

        >>> sim = pyrtl.Simulation()
        >>> sim.step(provided_inputs={"input": 0xab})
        >>> hex(sim.inspect("output"))
        '0xb'

    :param wirevector_or_integer: A :class:`WireVector` or :class:`int` to truncate.
    :param bitwidth: The length to which ``wirevector_or_integer`` should be truncated.

    :return: A truncated :class:`WireVector` or :class:`int`, depending on the input
             type.
    """
    if bitwidth < 1:
        msg = "bitwidth must be a positive integer"
        raise PyrtlError(msg)
    x = wirevector_or_integer
    try:
        return x.truncate(bitwidth)
    except AttributeError:
        return x & ((1 << bitwidth) - 1)


class MatchedFields(NamedTuple):
    """Result returned by :func:`match_bitpattern`."""

    matched: WireVector
    """1-bit :class:`WireVector` indicating if ``w`` matches ``bitpattern``."""
    fields: NamedTuple
    """:class:`NamedTuple` containing the matched fields, if any."""

    def __enter__(self):
        from pyrtl.conditional import _push_condition

        _push_condition(self.matched)
        return self.fields

    def __exit__(self, *execinfo):
        from pyrtl.conditional import _pop_condition

        _pop_condition()


def match_bitpattern(
    w: WireVector, bitpattern: str, field_map: dict[str, str] | None = None
) -> MatchedFields:
    """Returns a single-bit :class:`WireVector` that is ``1`` if and only if ``w``
    matches the ``bitpattern``, and a tuple containing the matched fields, if any.
    Compatible with the ``with`` statement.

    This function will compare a multi-bit :class:`WireVector` to a specified pattern of
    bits, where some of the pattern can be "wildcard" bits. If any of the ``1`` or ``0``
    values specified in the bitpattern fail to match the :class:`WireVector` during
    execution, a ``0`` will be produced, otherwise the value carried on the wire will be
    ``1``. The wildcard characters can be any other alphanumeric character, with
    characters other than ``?`` having special functionality (see below). The string
    must have length equal to the :class:`WireVector` specified, although whitespace and
    underscore characters will be ignored and can be used for pattern readability.

    For all other characters besides ``1``, ``0``, or ``?``, a tuple of
    :class:`WireVectors<WireVector>` will be returned as the second return value. Each
    character will be treated as the name of a field, and non-consecutive fields with
    the same name will be concatenated together, left-to-right, into a single field in
    the resultant tuple. For example, ``01aa1?bbb11a`` will match a string such as
    ``010010100111``, and the resultant matched fields are::

        (a, b) = (0b001, 0b100)

    where the ``a`` field is the concatenation of bits 9, 8, and 0, and the ``b`` field
    is the concenation of bits 5, 4, and 3. Thus, arbitrary characters beside ``?`` act
    as wildcard characters for the purposes of matching, with the additional benefit of
    returning the :class:`WireVectors<WireVector>` corresponding to those fields.

    A prime example of this is for decoding instructions. Here we decode some RISC-V::

        with pyrtl.conditional_assignment:
            with match_bitpattern(
                inst, "iiiiiiiiiiiirrrrr010ddddd0000011"
            ) as (imm, rs1, rd):
                regfile[rd] |= mem[(regfile[rs1] + imm.sign_extended(32)).truncate(32)]
                pc.next |= pc + 1
            with match_bitpattern(
                inst, "iiiiiiirrrrrsssss010iiiii0100011"
            ) as (imm, rs2, rs1):
                mem[(regfile[rs1] + imm.sign_extended(32)).truncate(32)] |= regfile[rs2]
                pc.next |= pc + 1
            with match_bitpattern(
                inst, "0000000rrrrrsssss111ddddd0110011"
            ) as (rs2, rs1, rd):
                regfile[rd] |= regfile[rs1] & regfile[rs2]
                pc.next |= pc + 1
            with match_bitpattern(
                inst, "0000000rrrrrsssss000ddddd0110011")
            as (rs2, rs1, rd):
                regfile[rd] |= (regfile[rs1] + regfile[rs2]).truncate(32)
                pc.next |= pc + 1
            # ...etc...

    Some smaller examples::

        # Basically the same as w == '0b0101'.
        m, _ = match_bitpattern(w, '0101')

        # m will be true when w is '0101' or '0111'.
        m, _ = match_bitpattern(w, '01?1')

        # m will be true when last two bits of w are '01'.
        m, _ = match_bitpattern(w, '??01')

        # spaces/underscores are ignored, same as line above.
        m, _ = match_bitpattern(w, '??_0 1')

        # All bits with the same letter make up same field.
        m, (a, b) = match_bitpattern(w, '01aa1?bbb11a')

        # Fields will be named `fs.foo` and `fs.bar`.
        m, fs = match_bitpattern(w, '01aa1?bbb11a', {'a': 'foo', 'b': 'bar'})

    :param w: The :class:`WireVector` to be compared to the ``bitpattern``
    :param bitpattern: A string holding the pattern (of bits and wildcards) to match
    :param field_map: (optional) A map from single-character field name in the
        bitpattern to the desired name of field in the returned :class:`NamedTuple`. If
        given, all non-``1``/``0``/``?`` characters in the ``bitpattern`` must be
        present in the map.

    :return: A :class:`NamedTuple` consisting of a 1-bit :class:`WireVector` carrying
             the result of the comparison, and a :class:`NamedTuple` containing the
             matched fields, if any.
    """
    w = as_wires(w)
    if not isinstance(bitpattern, str):
        msg = "bitpattern must be a string"
        raise PyrtlError(msg)
    bitpattern = bitpattern.replace("_", "").replace(" ", "")
    if len(w) != len(bitpattern):
        msg = "bitpattern string different length than wirevector provided"
        raise PyrtlError(msg)
    # Reverse ``bitpattern`` so index 0 is the least significant bit. This makes
    # ``w[i]`` and ``reversed_bitpattern[i]`` refer to the same bit ``i``.
    reversed_bitpattern = bitpattern[::-1]

    zero_bits = [w[index] for index, x in enumerate(reversed_bitpattern) if x == "0"]
    one_bits = [w[index] for index, x in enumerate(reversed_bitpattern) if x == "1"]
    match = rtl_all(*one_bits) & ~rtl_any(*zero_bits)

    def field_name(name: str) -> str:
        """Retrieve a field's name from ``field_map``."""
        if field_map is not None:
            if name not in field_map:
                msg = (
                    f"field_map argument has been given, but {name} field is not "
                    "present"
                )
                raise PyrtlError(msg)
            return field_map[name]
        return name

    # ``fields`` maps from ``field_name`` to a list of WireVectors that match
    # ``field_name``.
    fields = collections.defaultdict(list)
    for i, c in enumerate(reversed_bitpattern):
        if c not in "01?":
            fields[c].append(w[i])
    # Sort ``fields`` by each field's position in ``bitpattern`` and convert ``fields``
    # to a list of tuples.
    fields = sorted(fields.items(), key=lambda m: bitpattern.index(m[0]))
    Fields = collections.namedtuple("Fields", [field_name(name) for name, _ in fields])
    fields = Fields(
        **{
            field_name(name): concat_list(wirevector_list)
            for name, wirevector_list in fields
        }
    )

    return MatchedFields(match, fields)


def bitpattern_to_val(bitpattern: str, *ordered_fields, **named_fields) -> int:
    """Return an unsigned integer representation of field format filled with the
    provided values.

    This function will compare a specified pattern of bits, where some of the pattern
    can be "wildcard" bits. The wildcard bits must all be named with a single letter
    and, unlike the related function :func:`match_bitpattern`, no "?" can be used. The
    function will take the provided ``bitpattern`` and create an integer that
    substitutes the provided fields in for the given wildcards at the bit level. This
    sort of bit substitution is useful when creating values for testing when the
    resulting values will be "chopped" up by the hardware later (e.g. instruction decode
    or other bitfield heavy functions).

    If a special keyword argument, ``field_map``, is provided, then the named fields
    provided can be longer, human-readable field names, which will correspond to the
    field in the bitpattern according to the ``field_map``. See the third example below.

    Examples::

        >>> # RISC-V ADD instruction.
        >>> bin(bitpattern_to_val('0000000sssssrrrrr000ddddd0110011', s=2, r=1, d=3))
        '0b1000001000000110110011'

        >>> # RISC-V SW instruction.
        >>> bin(bitpattern_to_val('iiiiiiisssssrrrrr010iiiii0100011', i=1, s=4, r=3))
        '0b10000011010000010100011'

        >>> # RISC-V SW instruction.
        >>> bin(bitpattern_to_val(
        ...     'iiiiiiisssssrrrrr010iiiii0100011',
        ...         imm=1, rs2=4, rs1=3,
        ...         field_map={'i': 'imm', 's': 'rs2', 'r': 'rs1'}))
        '0b10000011010000010100011'

    :param bitpattern: A string holding the pattern (of bits and wildcards) to match
    :param ordered_fields: A list of parameters to be matched to the provided bit
        pattern in the order provided. If ``ordered_fields`` are provided then no
        ``named_fields`` can be used.
    :param named_fields: A list of parameters to be matched to the provided bit pattern
        by the names provided. If ``named_fields`` are provided then no
        ``ordered_fields`` can be used. A special keyword argument, ``field_map``, can
        be provided, which will allow you to specify a correspondence between the
        1-letter field names in the bitpattern string and longer, human readable field
        names. See the example above.

    :return: An unsigned integer carrying the result of the field substitution.
    """

    if not bitpattern:
        msg = "bitpattern must be nonempty"
        raise PyrtlError(msg)

    if len(ordered_fields) > 0 and len(named_fields) > 0:
        msg = "named and ordered fields cannot be mixed"
        raise PyrtlError(msg)

    def letters_in_field_order():
        seen = []
        for c in bitpattern:
            if c != "0" and c != "1" and c not in seen:
                seen.append(c)
        return seen

    field_map = None
    if "field_map" in named_fields:
        field_map = named_fields["field_map"]
        named_fields.pop("field_map")

    bitlist = []
    lifo = letters_in_field_order()
    if ordered_fields:
        if len(lifo) != len(ordered_fields):
            msg = "number of fields and number of unique patterns do not match"
            raise PyrtlError(msg)
        intfields = [int(f) for f in ordered_fields]
    else:
        if len(lifo) != len(named_fields):
            msg = "number of fields and number of unique patterns do not match"
            raise PyrtlError(msg)
        try:

            def fn(n):
                return field_map[n] if field_map else n

            intfields = [int(named_fields[fn(n)]) for n in lifo]
        except KeyError as exc:
            msg = f"bitpattern field {exc.args[0]} was not provided in named_field list"
            raise PyrtlError(msg) from exc

    fmap = dict(zip(lifo, intfields))
    for c in bitpattern[::-1]:
        if c == "0" or c == "1":
            bitlist.append(c)
        elif c == "?":
            msg = "all fields in the bitpattern must have names"
            raise PyrtlError(msg)
        else:
            bitlist.append(str(fmap[c] & 0x1))  # append lsb of the field
            fmap[c] = fmap[c] >> 1  # and bit shift by one position
    for f, intfield in fmap.items():
        if intfield not in [0, -1]:
            msg = f"too many bits given to value to fit in field {f}"
            raise PyrtlError(msg)
    if len(bitpattern) != len(bitlist):
        msg = "resulting values have different bitwidths"
        raise PyrtlInternalError(msg)
    final_str = "".join(bitlist[::-1])
    return int(final_str, 2)


def chop(w: WireVector, *segment_widths: int) -> list[WireVector]:
    """Returns a list of :class:`WireVectors<WireVector>`, each a slice of the original
    ``w``.

    This function chops a :class:`WireVector` into a set of smaller
    :class:`WireVectors<WireVector>` of different lengths. It is most useful when
    multiple "fields" are contained with a single :class:`WireVector`, for example when
    breaking apart an instruction.

    For example, if you wish to break apart a 32-bit MIPS I-type (Immediate) instruction
    you know it has an 6-bit opcode, 2 5-bit operands, and 16-bit offset. You could take
    each of those slices in absolute terms: ``offset=instr[0:16]``, ``rt=instr[16:21]``
    and so on, but then you have to do the arithmetic yourself. With this function you
    can do all the fields at once which can be seen in the examples below.

    As a check, ``chop`` will throw an error if the sum of the lengths of the fields
    given is not the same as the length of the :class:`WireVector` to ``chop``. Note
    also that ``chop`` assumes that the "rightmost" arguments are the least signficant
    bits (just like :func:`concat`) which is normal for hardware functions but makes the
    list order a little counter intuitive.

    .. note::

        Consider using :func:`wire_struct` or :func:`wire_matrix` instead, which helps
        with consistently disassembling, naming, and reassembling fields.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> input = pyrtl.Input(name="input", bitwidth=12)

        >>> high, middle, low = pyrtl.chop(input, 4, 4, 4)
        >>> high.name = "high"
        >>> middle.name = "middle"
        >>> low.name = "low"

        >>> sim = pyrtl.Simulation()
        >>> sim.step(provided_inputs={"input": 0xabc})

        >>> hex(sim.inspect("high"))
        '0xa'
        >>> hex(sim.inspect("middle"))
        '0xb'
        >>> hex(sim.inspect("low"))
        '0xc'

    :param w: The :class:`WireVector` to be chopped up into segments
    :param segment_widths: Additional arguments are integers which are bitwidths

    :return: A list of :class:`WireVectors<WireVector>` each with a proper segment
             width.
    """
    w = as_wires(w)
    for seg in segment_widths:
        if not isinstance(seg, int):
            msg = "segment widths must be integers"
            raise PyrtlError(msg)
    if sum(segment_widths) != len(w):
        msg = "sum of segment widths must equal length of wirevetor"
        raise PyrtlError(msg)

    n_segments = len(segment_widths)
    starts = [sum(segment_widths[i + 1 :]) for i in range(n_segments)]
    ends = [sum(segment_widths[i:]) for i in range(n_segments)]
    return [w[s:e] for s, e in zip(starts, ends)]


def input_list(
    names: str | list[str], bitwidth: int | list[int] | None = None
) -> list[Input]:
    """Allocate and return a list of :class:`Inputs<Input>`.

    See :func:`wirevector_list`. Equivalent to::

        wirevector_list(names, bitwidth, wvtype=pyrtl.Input)

    .. WARNING::

        Avoid using this function. Lists of ``Inputs`` can be created with list
        comprehensions, which are easier to understand because they compose familiar
        concepts, rather than introducing a new concept.

    :param names: Names for the ``Inputs``. Can be a list or single
        comma/space-separated string
    :param bitwidth: The desired bitwidth for the resulting ``Inputs``.
    """
    return wirevector_list(names, bitwidth, wvtype=Input)


def output_list(
    names: str | list[str], bitwidth: int | list[int] | None = None
) -> list[Output]:
    """Allocate and return a list of :class:`Outputs<Output>`.

    See :func:`wirevector_list`. Equivalent to::

        wirevector_list(names, bitwidth, wvtype=pyrtl.Output)

    .. WARNING::

        Avoid using this function. Lists of ``Outputs`` can be created with list
        comprehensions, which are easier to understand because they compose familiar
        concepts, rather than introducing a new concept.

    :param names: Names for the ``Outputs``. Can be a list or single
        comma/space-separated string
    :param bitwidth: The desired bitwidth for the resulting ``Outputs``.
    """
    return wirevector_list(names, bitwidth, wvtype=Output)


def register_list(
    names: str | list[str], bitwidth: int | list[int] | None = None
) -> list[Register]:
    """Allocate and return a list of :class:`Registers<Register>`.

    See :func:`wirevector_list`. Equivalent to::

        wirevector_list(names, bitwidth, wvtype=pyrtl.Register)

    .. WARNING::

        Avoid using this function. Lists of ``Registers`` can be created with list
        comprehensions, which are easier to understand because they compose familiar
        concepts, rather than introducing a new concept.

    :param names: Names for the ``Registers``. Can be a list or single
        comma/space-separated string
    :param bitwidth: The desired bitwidth for the resulting ``Registers``.
    """
    return wirevector_list(names, bitwidth, wvtype=Register)


def wirevector_list(
    names: str | list[str],
    bitwidth: int | list[int] | None = None,
    wvtype: type[WireVector] = WireVector,
) -> list[WireVector]:
    """Allocate and return a list of :class:`WireVectors<WireVector>`.

    The strings in ``names`` can also contain an additional bitwidth specification,
    separated by a ``/``. This cannot be used in combination with a
    :attr:`~WireVector.bitwidth` value other than ``1``.

    Examples::

        wirevector_list(['name1', 'name2', 'name3'])
        wirevector_list('name1, name2, name3')
        wirevector_list('input1 input2 input3', bitwidth=8, wvtype=pyrtl.Input)
        wirevector_list('output1, output2 output3', bitwidth=3, wvtype=pyrtl.Output)
        wirevector_list('two_bits/2, four_bits/4, eight_bits/8')
        wirevector_list(['name1', 'name2', 'name3'], bitwidth=[2, 4, 8])

    .. WARNING::

        Avoid using this function. Create lists of :class:`WireVectors<WireVector>` with
        list comprehensions, which are easier to understand because they compose
        familiar concepts, rather than introducing a new concept::

            [WireVector(name) for name in ['name1', 'name2', 'name3']]
            [Input(name) for name in 'input1 input2 input3'.split(' ')]

    :param names: Names for the :class:`WireVectors<WireVector>`. Can be a list or a
        single comma/space-separated string
    :param bitwidth: The desired bitwidth for the resulting
        :class:`WireVectors<WireVector>`.
    :param wvtype: The :class:`WireVector` type to create.
    """
    if isinstance(names, str):
        names = names.replace(",", " ").split()

    if any("/" in name for name in names) and bitwidth is not None:
        msg = 'only one of optional "/" or bitwidth parameter allowed'
        raise PyrtlError(msg)

    if bitwidth is None:
        bitwidth = 1
    if isinstance(bitwidth, numbers.Integral):
        bitwidth = [bitwidth] * len(names)
    if len(bitwidth) != len(names):
        raise ValueError(
            "number of names "
            + str(len(names))
            + " should match number of bitwidths "
            + str(len(bitwidth))
        )

    wirelist = []
    for fullname, bw in zip(names, bitwidth):
        try:
            name, bw = fullname.split("/")
        except ValueError:
            name = fullname
        wirelist.append(wvtype(bitwidth=int(bw), name=name))
    return wirelist


def val_to_signed_integer(value: int, bitwidth: int) -> int:
    """Return ``value`` interpreted as a two's complement signed integer.

    Reinterpret an unsigned integer (not a :class:`WireVector`!) as a signed integer.
    This is useful for printing and interpreting two's complement values::

        >>> val_to_signed_integer(0xff, bitwidth=8)
        -1

    ``val_to_signed_integer`` can also be used as an ``repr_func`` for
    :meth:`SimulationTrace.render_trace`, to display signed integers in traces::

        bitwidth = 3
        counter = Register(name='counter', bitwidth=bitwidth)
        counter.next <<= counter + 1
        sim = Simulation()
        sim.step_multiple(nsteps=2 ** bitwidth)

        # Generates a trace like:
        #        │0 │1 │2 │3 │4 │5 │6 │7
        #
        # counter ──┤1 │2 │3 │-4│-3│-2│-1
        sim.tracer.render_trace(repr_func=val_to_signed_integer)

    :func:`infer_val_and_bitwidth` performs the opposite conversion::

        >>> integer = val_to_signed_integer(0xff, bitwidth=8)
        >>> hex(infer_val_and_bitwidth(integer, bitwidth=8).value)
        '0xff'

    :param value: A Python integer holding the value to convert.
    :param bitwidth: The length of the integer in bits to assume for conversion.

    :return: ``value`` as a signed integer
    """
    if isinstance(value, WireVector) or isinstance(bitwidth, WireVector):
        msg = "inputs must not be wirevectors"
        raise PyrtlError(msg)
    if bitwidth < 1:
        msg = "bitwidth must be a positive integer"
        raise PyrtlError(msg)

    neg_mask = 1 << (bitwidth - 1)
    neg_part = value & neg_mask

    pos_mask = neg_mask - 1
    pos_part = value & pos_mask

    return pos_part - neg_part


def formatted_str_to_val(data: str, format: str, enum_set=None) -> int:
    """Return an unsigned integer representation of ``data`` in a specified ``format``.

    Given a string (not a :class:`WireVector`!) convert that to an unsigned integer
    ready for input to the simulation environment. This helps deal with signed/unsigned
    numbers (simulation assumes the values have been converted via two's complement
    already), but it also takes hex, binary, and enum types as inputs. It is easiest to
    see how it works with some examples::

        >>> formatted_str_to_val('2', 's3')
        2
        >>> bin(formatted_str_to_val('-1', 's3'))
        '0b111'
        >>> bin(formatted_str_to_val('101', 'b3'))
        '0b101'
        >>> formatted_str_to_val('5', 'u3')
        5
        >>> bin(formatted_str_to_val('-3', 's3'))
        '0b101'
        >>> formatted_str_to_val('a', 'x3')
        10

        >>> from enum import IntEnum
        >>> class Ctl(IntEnum):
        ...     ADD = 5
        ...     SUB = 12
        >>> formatted_str_to_val('ADD', 'e3/Ctl', [Ctl])
        5
        >>> formatted_str_to_val('SUB', 'e3/Ctl', [Ctl])
        12

    :func:`val_to_formatted_str` performs the opposite conversion.

    :param data: A string holding the value to convert.
    :param format: A string holding a format which will be used to convert the data
        string.
    :param enum_set: An iterable of :class:`~enum.IntEnum` which are used as part of the
        conversion process.

    :return: ``data`` as a signed integer
    """
    type = format[0]
    bitwidth = int(format[1:].split("/")[0])
    bitmask = (1 << bitwidth) - 1
    if type == "s":
        rval = int(data) & bitmask
    elif type == "x":
        rval = int(data, 16)
    elif type == "b":
        rval = int(data, 2)
    elif type == "u":
        rval = int(data)
        if rval < 0:
            msg = "unsigned format requested, but negative value provided"
            raise PyrtlError(msg)
    elif type == "e":
        enumname = format.split("/")[1]
        enum_inst_list = [e for e in enum_set if e.__name__ == enumname]
        if len(enum_inst_list) == 0:
            msg = f'enum "{enumname}" not found in passed enum_set "{enum_set}"'
            raise PyrtlError(msg)
        rval = getattr(enum_inst_list[0], data).value
    else:
        msg = f"unknown format type {format}"
        raise PyrtlError(msg)
    return rval


def val_to_formatted_str(val: int, format: str, enum_set=None) -> str:
    """Return a string representation of the value given format specified.

    Given an unsigned integer (not a :class:`WireVector`!) convert that to a
    human-readable string. This helps deal with signed/unsigned numbers (simulation
    operates on values that have been converted via two's complement), but it also
    generates hex, binary, and enum types as outputs. It is easiest to see how it works
    with some examples::

        >>> val_to_formatted_str(2, 's3')
        '2'
        >>> val_to_formatted_str(7, 's3')
        '-1'
        >>> val_to_formatted_str(5, 'b3')
        '101'
        >>> val_to_formatted_str(5, 'u3')
        '5'
        >>> val_to_formatted_str(5, 's3')
        '-3'
        >>> val_to_formatted_str(10, 'x3')
        'a'

        >>> from enum import IntEnum
        >>> class Ctl(IntEnum):
        ...     ADD = 5
        ...     SUB = 12
        >>> val_to_formatted_str(5, 'e3/Ctl', [Ctl])
        'ADD'
        >>> val_to_formatted_str(12, 'e3/Ctl', [Ctl])
        'SUB'

    :func:`formatted_str_to_val` performs the opposite conversion.

    :param val: An unsigned integer to convert.
    :param format: A string holding a format which will be used to convert the data
        string.
    :param enum_set: An iterable of :class:`~enum.IntEnum` which are used as part of the
        conversion process.

    :return: a human-readable string representing `val`.
    """
    type = format[0]
    bitwidth = int(format[1:].split("/")[0])
    if type == "s":
        rval = str(val_to_signed_integer(val, bitwidth))
    elif type == "x":
        rval = hex(val)[2:]  # cuts off '0x' at the start
    elif type == "b":
        rval = bin(val)[2:]  # cuts off '0b' at the start
    elif type == "u":
        rval = str(int(val))  # nothing fancy
    elif type == "e":
        enumname = format.split("/")[1]
        enum_inst_list = [e for e in enum_set if e.__name__ == enumname]
        if len(enum_inst_list) == 0:
            msg = f'enum "{enumname}" not found in passed enum_set "{enum_set}"'
            raise PyrtlError(msg)
        rval = enum_inst_list[0](val).name
    else:
        msg = f"unknown format type {format}"
        raise PyrtlError(msg)
    return rval


class ValueBitwidthTuple(NamedTuple):
    """Return type for :func:`infer_val_and_bitwidth`."""

    value: int
    """Inferred value."""
    bitwidth: int
    """Inferred bitwidth."""


def infer_val_and_bitwidth(
    rawinput: int | bool | str, bitwidth: int | None = None, signed: bool = False
) -> ValueBitwidthTuple:
    """Return a ``(value, bitwidth)`` :class:`tuple` inferred from the specified input.

    Given a boolean, integer, or Verilog-style string constant, this function returns a
    :class:`ValueBitwidthTuple` ``(value, bitwidth)`` which are inferred from the
    specified ``rawinput``. If ``signed`` is ``True``, bits will be included to ensure a
    proper two's complement representation is possible, otherwise it assumes a standard
    unsigned representation. Error checks are performed that determine if the bitwidths
    specified are sufficient and appropriate for the values specified. Examples::

        >>> infer_val_and_bitwidth(2, bitwidth=5)
        ValueBitwidthTuple(value=2, bitwidth=5)

        >>> # Infer bitwidth from value.
        >>> infer_val_and_bitwidth(3)
        ValueBitwidthTuple(value=3, bitwidth=2)
        >>> infer_val_and_bitwidth(3).bitwidth
        2

        >>> # Signed values need an additional sign bit.
        >>> infer_val_and_bitwidth(3, signed=True)
        ValueBitwidthTuple(value=3, bitwidth=3)

        >>> val, bitwidth = infer_val_and_bitwidth(-1, bitwidth=3)
        >>> (bin(val), bitwidth)
        ('0b111', 3)

        >>> infer_val_and_bitwidth("5'd12")
        ValueBitwidthTuple(value=12, bitwidth=5)

    :func:`val_to_signed_integer` performs the opposite conversion::

        >>> val, bitwidth = infer_val_and_bitwidth(-1, bitwidth=3)
        >>> val_to_signed_integer(val, bitwidth)
        -1

    :param rawinput: a bool, int, or Verilog-style string constant
    :param bitwidth: an integer bitwidth or (by default) ``None``
    :param signed: a bool (by default ``False``) to include bits for proper two's
        complement

    :return: tuple of integers ``(value, bitwidth)``
    """

    if isinstance(rawinput, bool):
        return _convert_bool(rawinput, bitwidth, signed)
    if isinstance(rawinput, numbers.Integral):
        return _convert_int(rawinput, bitwidth, signed)
    if isinstance(rawinput, str):
        return _convert_verilog_str(rawinput, bitwidth, signed)
    msg = (
        f'error, the value provided is of an improper type, "{type(rawinput)}" proper '
        "types are bool, int, and string"
    )
    raise PyrtlError(msg)


def _convert_bool(
    bool_val: bool, bitwidth: int | None = None, signed: bool = False
) -> ValueBitwidthTuple:
    if signed:
        msg = "error, booleans cannot be signed (convert to int first)"
        raise PyrtlError(msg)
    num = int(bool_val)
    if bitwidth is None:
        bitwidth = 1
    if bitwidth != 1:
        msg = "error, boolean has bitwidth not equal to 1"
        raise PyrtlError(msg)
    return ValueBitwidthTuple(num, bitwidth)


def _convert_int(
    val: numbers.Integral, bitwidth: int | None = None, signed: bool = False
) -> ValueBitwidthTuple:
    # Convert val from numbers.Integral to int. This avoids issues with
    # limited-precision types like numpy.int32.
    val = int(val)

    if val >= 0:
        num = val
        # infer bitwidth if it is not specified explicitly
        min_bitwidth = (
            len(bin(num)) - 2
        )  # the -2 for the "0b" at the start of the string
        if signed and val != 0:
            min_bitwidth += 1  # extra bit needed for the zero

        if bitwidth is None:
            bitwidth = min_bitwidth
        elif bitwidth < min_bitwidth:
            msg = (
                f"bitwidth specified ({bitwidth}) is insufficient to represent "
                f"constant {val}"
            )
            raise PyrtlError(msg)

    else:  # val is negative
        if not signed and bitwidth is None:
            msg = (
                f"negative constant {val} requires either signed=True or specified "
                "bitwidth"
            )
            raise PyrtlError(msg)

        if bitwidth is None:
            bitwidth = 1 if val == -1 else len(bin(~val)) - 1

        if (val >> bitwidth - 1) != -1:
            msg = f"insufficient bits ({bitwidth}) for negative number {val}"
            raise PyrtlError(msg)

        num = val & ((1 << bitwidth) - 1)  # result is a two's complement value
    return ValueBitwidthTuple(num, bitwidth)


def _convert_verilog_str(
    val: str, bitwidth: int | None = None, signed: bool = False
) -> ValueBitwidthTuple:
    if signed:
        msg = 'error, "signed" option with Verilog-style string constants not supported'
        raise PyrtlError(msg)

    bases = {"b": 2, "o": 8, "d": 10, "h": 16, "x": 16}

    neg = False
    if val.startswith("-"):
        neg = True
        val = val[1:]

    split_string = val.lower().split("'")
    if len(split_string) != 2:
        msg = "error, string not in Verilog style format"
        raise PyrtlError(msg)
    try:
        verilog_bitwidth = int(split_string[0])
        bitwidth = (
            bitwidth or verilog_bitwidth
        )  # if bitwidth is None, use verilog_bitwidth
        if verilog_bitwidth > bitwidth:
            msg = (
                f"bitwidth parameter passed ({bitwidth}) cannot fit Verilog-style "
                f"constant with bitwidth {verilog_bitwidth} (if bitwidth=None is used, "
                "PyRTL will determine the bitwidth from the Verilog-style constant "
                "specification)"
            )
            raise PyrtlError(msg)

        sval = split_string[1]
        if sval[0] == "s":
            msg = "error, signed integers are not supported in Verilog-style constants"
            raise PyrtlError(msg)
        base = 10
        if sval[0] in bases:
            base = bases[sval[0]]
            sval = sval[1:]
        sval = sval.replace("_", "")
        num = int(sval, base)
    except (IndexError, ValueError) as exc:
        msg = "error, string not in Verilog style format"
        raise PyrtlError(msg) from exc

    if neg and num:
        if num >> bitwidth - 1:
            msg = "error, insufficient bits for negative number"
            raise PyrtlError(msg)
        num = (1 << bitwidth) - num

    if num >> bitwidth != 0:
        msg = (
            f"specified bitwidth {bitwidth} for Verilog constant insufficient to store "
            f"value {num}"
        )
        raise PyrtlError(msg)

    return ValueBitwidthTuple(num, bitwidth)


def get_stacks(*wires):
    call_stack = getattr(wires[0], "init_call_stack", None)
    if not call_stack:
        return (
            "    No call info found for wires: use set_debug_mode() to provide more "
            "information\n"
        )
    return "\n".join(str(wire) + ":\n" + get_stack(wire) for wire in wires)


def get_stack(wire):
    if not isinstance(wire, WireVector):
        msg = "Only WireVectors can be traced"
        raise PyrtlError(msg)

    call_stack = getattr(wire, "init_call_stack", None)
    if call_stack:
        frames = " ".join(frame for frame in call_stack[:-1])
        return "Wire Traceback, most recent call last \n" + frames + "\n"
    return (
        "    No call info found for wire: use set_debug_mode() to provide more "
        "information"
    )


def _check_for_loop(block=None):
    block = working_block(block)
    logic_left = block.logic.copy()
    wires_left = block.wirevector_subset(exclude=(Input, Const, Output, Register))
    prev_logic_left = len(logic_left) + 1
    while prev_logic_left > len(logic_left):
        prev_logic_left = len(logic_left)
        nets_to_remove = (
            set()
        )  # bc it's not safe to mutate a set inside its own iterator
        for net in logic_left:
            if not any(n_wire in wires_left for n_wire in net.args):
                nets_to_remove.add(net)
                wires_left.difference_update(net.dests)
        logic_left -= nets_to_remove

    if len(logic_left) == 0:
        return None
    return wires_left, logic_left


def find_loop(block=None):
    block = working_block(block)
    block.sanity_check()  # make sure that the block is sane first

    result = _check_for_loop(block)
    if not result:
        return None
    wires_left, logic_left = result

    class _FilteringState:
        def __init__(self, dst_w):
            self.dst_w = dst_w
            self.arg_num = -1

    def dead_end():
        # clean up after a wire is found to not be part of the loop
        wires_left.discard(cur_item.dst_w)
        current_wires.discard(cur_item.dst_w)
        del checking_stack[-1]

    # now making a map to quickly look up nets
    dest_nets = {dest_w: net_ for net_ in logic_left for dest_w in net_.dests}
    initial_w = random.sample(list(wires_left), 1)[0]

    current_wires = set()
    checking_stack = [_FilteringState(initial_w)]

    # we don't use a recursive method as Python has a limited stack (default: 999
    # frames)
    while checking_stack:
        cur_item = checking_stack[-1]
        if cur_item.arg_num == -1:
            #  first time testing this item
            if cur_item.dst_w not in wires_left:
                dead_end()
                continue
            current_wires.add(cur_item.dst_w)
            cur_item.net = dest_nets[cur_item.dst_w]
            if cur_item.net.op == "r":
                dead_end()
                continue
        cur_item.arg_num += 1  # go to the next item
        if cur_item.arg_num == len(cur_item.net.args):
            dead_end()
            continue
        next_wire = cur_item.net.args[cur_item.arg_num]
        if next_wire not in current_wires:
            current_wires.add(next_wire)
            checking_stack.append(_FilteringState(next_wire))
        else:  # We have found the loop!!!!!
            loop_info = []
            for f_state in reversed(checking_stack):
                loop_info.append(f_state)
                if f_state.dst_w is next_wire:
                    break
            else:
                msg = "Shouldn't get here! Couldn't figure out the loop"
                raise PyrtlError(msg)
            return loop_info
    msg = "Error in detecting loop"
    raise PyrtlError(msg)


def find_and_print_loop(block=None):
    loop_data = find_loop(block)
    print_loop(loop_data)
    return loop_data


def print_loop(loop_data):
    if not loop_data:
        print("No Loop Found")
    else:
        print("Loop found:")
        print("\n".join(f"{fs.net}" for fs in loop_data))
        print()


def _currently_in_jupyter_notebook():
    """Return true if running under Jupyter notebook, otherwise return False.

    We want to check for more than just the presence of __IPYTHON__ because that is
    present in both Jupyter notebooks and IPython terminals.
    """
    try:
        # get_ipython() is in the global namespace when ipython is started
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        if shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        return False  # Other type
    except NameError:
        return False  # Probably standard Python interpreter


def _print_netlist_latex(netlist):
    """Print each net in netlist in a Latex array"""
    from IPython.display import Latex, display

    out = "\n\\begin{array}{ \\| c \\| c \\| l \\| }\n"
    out += "\n\\hline\n"
    out += "\\hline\n".join(str(n) for n in netlist)
    out += "\\hline\n\\end{array}\n"
    display(Latex(out))


class _NetCount:
    """Helper class to track when to stop an iteration that depends on number of nets

    Mainly useful for iterations that are for optimization
    """

    def __init__(self, block=None):
        self.block = working_block(block)
        self.prev_nets = len(self.block.logic) * 1000

    def shrank(self, block=None, percent_diff=0, abs_diff=1):
        """Returns whether a block has fewer nets than before

        This function checks whether the change in the number of nets is greater than
        the percentage and absolute difference thresholds.

        :param Block block: block to check (if changed)
        :param Number percent_diff: percentage difference threshold
        :param int abs_diff: absolute difference threshold

        :return: boolean
        """
        if block is None:
            block = self.block
        cur_nets = len(block.logic)
        net_goal = self.prev_nets * (1 - percent_diff) - abs_diff
        less_nets = cur_nets <= net_goal
        self.prev_nets = cur_nets
        return less_nets

    shrinking = shrank


# _ComponentMeta holds the component's name, bitwidth, and type. If the _ComponentMeta's
# type is None, then the default component_type should be used instead.
_ComponentMeta = collections.namedtuple("_ComponentMeta", ["name", "bitwidth", "type"])


def _make_component(
    component_meta: _ComponentMeta,
    block: Block,
    name: str,
    component_type,
    component_value,
):
    """Determine the component's type, instantiate it, and set its value."""
    # Determine the component's actual type.
    #
    # If the _ComponentMeta specifies a type, then the component is a wire_struct or a
    # wire_matrix. The _ComponentMeta's type must be used as the component's primary
    # type, and the default_component_type becomes the component's concatenated_type.
    #
    # If the _ComponentMeta does not specify a type, the component uses the default
    # component_type.
    if component_meta.type is None:
        actual_component_type = component_type
    else:
        actual_component_type = component_meta.type

    component_name = ""
    if len(name) > 0:
        if isinstance(component_meta.name, str):
            # wire_struct components are named with strings and printed with
            # dots, like `struct.component`.
            component_name = name + "." + component_meta.name
        else:
            # wire_matrix components are numbered with integers and printed
            # with brackets, like `matrix[0]`.
            component_name = name + "[" + str(component_meta.name) + "]"

    # The logic below always creates a new wire_struct, wire_matrix, or WireVector for
    # each component. If the component_value already has the appropriate type and name,
    # we could use the component_value directly and we don't need a new
    # struct/matrix/Vector. Correctly detecting these opportunities is complicated, so
    # we keep things simple for now.
    #
    # Components are always initialized with one concatenated component_value, which
    # provides values for all its wires. This implies that component wire_structs and
    # wire_matricies always call _split().
    if hasattr(actual_component_type, "_is_wire_struct"):
        # Make a wire_struct component. component_value may be None.
        component_kwargs = {actual_component_type._class_name: component_value}
        component = actual_component_type(
            name=component_name,
            block=block,
            concatenated_type=component_type,
            **component_kwargs,
        )
    elif hasattr(actual_component_type, "_is_wire_matrix"):
        # Make a wire_matrix component. component_value may be None.
        component = actual_component_type(
            name=component_name,
            block=block,
            concatenated_type=component_type,
            values=[component_value],
        )
    elif isinstance(component_value, int) and actual_component_type is WireVector:
        # Special case: simplify the component type to Const.
        component = Const(
            bitwidth=component_meta.bitwidth,
            name=component_name,
            block=block,
            val=component_value,
        )
    else:
        # Make a WireVector component.
        component = actual_component_type(
            bitwidth=component_meta.bitwidth, name=component_name, block=block
        )
        if component_value is not None:
            component <<= component_value

    return component


def _bitslice(value: int, start: int, end: int) -> int:
    """Slice an integer value bitwise, from start to end."""
    mask = (1 << (end - start)) - 1
    return (value >> start) & mask


def _slice(
    block: Block,
    schema: list[_ComponentMeta],
    bitwidth: int,
    component_type: type,
    name: str,
    concatenated,
    components,
    concatenated_value,
):
    """Slice ``concatenated`` into components.

    ``concatenated_value`` is the driver for ``concatenated``. Some optimizations are
    possible by inspecting ``concatenated_value``, for example we immediately slice
    Consts rather than generating slicing logic.
    """
    if concatenated_value is not None and not isinstance(concatenated, Const):
        concatenated <<= concatenated_value

    end_index = bitwidth
    for component_meta in schema:
        if isinstance(concatenated_value, int):
            # Special case: immediately slice Const values.
            component_value = _bitslice(
                concatenated_value, end_index - component_meta.bitwidth, end_index
            )
        else:
            component_value = concatenated[
                end_index - component_meta.bitwidth : end_index
            ]

        end_index -= component_meta.bitwidth

        component = _make_component(
            component_meta=component_meta,
            block=block,
            name=name,
            component_type=component_type,
            component_value=component_value,
        )

        components[component_meta.name] = component


def _concatenate(
    block: Block,
    schema: list[_ComponentMeta],
    component_type: type,
    name: str,
    concatenated,
    components,
    component_map,
):
    """Concatenate components from ``component_map`` to ``concatenated``."""
    all_components = []
    for component_meta in schema:
        component_value = component_map[component_meta.name]

        component = _make_component(
            component_meta=component_meta,
            block=block,
            name=name,
            component_type=component_type,
            component_value=component_value,
        )

        components[component_meta.name] = component
        all_components.append(component)

    concatenated <<= concat(*all_components)


def wire_struct(wire_struct_spec):
    """Decorator that assigns names to :class:`WireVector` slices.

    ``@wire_struct`` assigns names to *non-overlapping* :class:`WireVector` slices.
    Suppose we have an 8-bit wide :class:`WireVector` called ``byte``. We can refer to
    all 8 bits with the name ``byte``, but ``@wire_struct`` lets us refer to slices by
    name, for example we could name the high 4 bits ``byte.high`` and the low 4 bits
    ``byte.low``. Without ``@wire_struct``, we would refer to these slices as
    ``byte[4:8]`` and ``byte[0:4]``, which are prone to off-by-one errors and harder to
    read.

    .. note::

        See `example1.2-wire-struct.py
        <https://github.com/UCSBarchlab/PyRTL/blob/development/examples/example1.2-wire-struct.py>`_
        for additional examples.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    The example ``Byte`` ``@wire_struct`` can be defined as::

        >>> @wire_struct
        ... class Byte:
        ...     high: 4
        ...     low: 4

    Construction
    ------------

    Once a ``@wire_struct`` class is defined, it can be instantiated by providing
    drivers for all of its wires. This can be done in two ways:

    1. Provide a driver for *each* component wire, for example::

            >>> byte = Byte(high=0xA, low=0xB)

       Note how the component names (``high``, ``low``) are used as keyword args for the
       constructor. Drivers must be provided for *all* components.

    2. Provide a driver for the entire ``@wire_struct``, for example::

            >>> byte = Byte(Byte=0xAB)

       Note how the class name (``Byte``) is used as a keyword arg for the constructor.

    Accessing Slices
    ----------------

    After instantiating a ``@wire_struct``, the instance functions as a
    :class:`WireVector` containing all the wires. For example, ``byte`` functions as a
    :class:`WireVector` with bitwidth 8::

        >>> byte = Byte(Byte=0xAB)
        >>> byte.bitwidth
        8

    The named slice can be accessed through the ``.`` operator (``__getattr__``), for
    example ``byte.high`` and ``byte.low``, which both function as :class:`WireVector`
    with bitwidth 4::

        >>> byte = Byte(Byte=0xAB)
        >>> byte.high.bitwidth
        4
        >>> byte.low.bitwidth
        4

    Both the instance and the slices are first-class :class:`WireVector`, so they can be
    manipulated with all the usual PyRTL operators.

    ``len()`` returns the number of components in the ``@wire_struct``::

        >>> byte = Byte(Byte=0xAB)
        >>> len(byte)
        2

    Naming
    ------

    A ``@wire_struct`` can be assigned a name in the usual way::

        >>> byte = Byte(name="my_byte", high=0xA, low=0xB)

    When a ``@wire_struct`` is assigned a name (``my_byte``), its components will be
    assigned dotted names (``my_byte.high``, ``my_byte.low``)::

        >>> byte.high.name
        'my_byte.high'
        >>> byte.low.name
        'my_byte.low'

    .. WARNING::

        All ``@wire_struct`` names are only set during construction. You can later
        rename a ``@wire_struct`` or its components, but those changes are local, and
        will not propagate to other ``@wire_struct`` components.

    Composition
    -----------

    .. doctest comment::

        These `Composition` examples are currently not ``doctested`` because the
        ``doctest`` environment converts all class attributes to strings, due to its use
        of ``exec``. In the example below, the attribute will be "Byte" (``str``),
        instead of ``Byte`` (``class``), which breaks ``wire_struct``.

    ``@wire_struct`` can be composed with itself, and with :func:`wire_matrix`. For
    example, we can define a ``Pixel`` that contains three ``Bytes``::

        @pyrtl.wire_struct
        class Pixel:
            red: Byte
            green: Byte
            blue: Byte

    Drivers must be specified for all components, but they can be specified at any
    level. All these examples construct an equivalent ``@wire_struct``::

        pixel = Pixel(Pixel=0xABCDEF)
        pixel = Pixel(red=0xAB, green=0xCD, blue=0xEF)
        pixel = Pixel(red=Byte(high=0xA, low=0xB), green=0xCD, blue=0xEF)
        pixel = Pixel(red=Byte(high=0xA, low=0xB),
                      green=Byte(high=0xC, low=0xD),
                      blue=0xEF)

    Hierarchical ``@wire_struct`` components are accessed by composing ``.`` operators::

        pixel
        pixel.red
        pixel.red.high
        pixel.red.low
        pixel.green
        pixel.green.high
        pixel.green.low
        pixel.blue
        pixel.blue.high
        pixel.blue.low

    ``@wire_struct`` can be composed with :func:`wire_matrix`::

        Word = pyrtl.wire_matrix(component_schema=8, size=4)

        @pyrtl.wire_struct
        class CacheLine:
            address: Word
            data: Word
            valid: 1

        cache_line = CacheLine(address=0x01234567, data=0x89ABCDEF, valid=1)

    Leaf-level components can be accessed by combining the ``.`` and ``[]`` operators,
    for example ``cache_line.address[3]``.

    Types
    -----

    You can change the type of a ``@wire_struct``'s components to a :class:`WireVector`
    subclass like :class:`Input` or :class:`Output` with the ``component_type``
    constructor argument::

        # Generates Outputs named ``output_byte.low`` and ``output_byte.high``.
        >>> byte = Byte(name="output_byte", component_type=pyrtl.Output,
        ...             Byte=0xCD)

        >>> isinstance(byte.high, pyrtl.Output)
        True
        >>> byte.high.name
        'output_byte.high'

    You can also change the type of the ``@wire_struct`` itself with the
    ``concatenated_type`` constructor argument::

        # Generates an Input named ``input_byte``.
        >>> input_byte = Byte(name="input_byte", concatenated_type=pyrtl.Input)

    .. NOTE::

        No values are specified for ``input_byte`` because its value is not known until
        simulation time.
    """
    # Convert the decorated class' annotations (dict of attr_name: attr_value)
    # to a list of _ComponentMetas.
    #
    # dict iteration order is guaranteed to be insertion order in Python 3.7+.
    schema = []
    for component_name, component_bitwidth in wire_struct_spec.__annotations__.items():
        # This is a hack for doctests, which convert all __annotations__ to str due to
        # its use of `exec`.
        if isinstance(component_bitwidth, str):
            component_bitwidth = int(component_bitwidth)

        if isinstance(component_bitwidth, int):
            # An ordinary component ("foo: 4") that should use the default
            # component_type.
            schema.append(
                _ComponentMeta(
                    name=component_name, bitwidth=component_bitwidth, type=None
                )
            )
        else:
            # A nested component ("bar: Byte") that must use the nested
            # component's type.
            schema.append(
                _ComponentMeta(
                    name=component_name,
                    bitwidth=component_bitwidth._bitwidth,
                    type=component_bitwidth,
                )
            )

    total_bitwidth = sum([component.bitwidth for component in schema])

    # Name of the decorated class.
    class_name = wire_struct_spec.__name__

    class _WireStruct(WrappedWireVector):
        """``wire_struct`` implementation: Concatenate or slice :class:`WireVector`.

        ``wire_struct`` works by either concatenating component :class:`WireVector`
        to create the ``wire_struct``'s full value, *or* slicing a
        ``wire_struct``s value to create component :class:`WireVectors<WireVector>`. A
        ``wire_struct`` can only concatenate or slice, not both. The decision
        to concatenate or slice is made in __init__.

        """

        _bitwidth = total_bitwidth
        _class_name = class_name
        _is_wire_struct = True

        def __init__(
            self,
            name="",
            block=None,
            concatenated_type=WireVector,
            component_type=WireVector,
            **kwargs,
        ):
            """Concatenate or slice :class:`WireVector` components.

            The remaining keyword args specify values for all wires. If the concatenated
            value is provided, its value must be provided with the keyword arg matching
            the decorated class name. For example, if the decorated class is::

                @wire_struct
                class Byte:
                    high: 4  # high is the 4 most significant bits.
                    low: 4   # low is the 4 least significant bits.

            then the concatenated value must be provided like this::

                byte = Byte(Byte=0xAB)

            And if the component values are provided instead, their values are set by
            keyword args matching the component names::

                byte = Byte(low=0xA, high=0xB)

            :param str name: The name of the concatenated wire. Must be unique. If none
                is provided, one will be autogenerated. If a name is provided,
                components will be assigned names of the form "{name}.{component_name}".
            :param Block block: The block containing the concatenated and component
                wires. Defaults to the :ref:`working_block`.
            :param type concatenated_type: Type for the concatenated
                :class:`WireVector`.
            :param type component_type: Type for each component.
            """
            # The concatenated WireVector contains all the _WireStruct's wires.
            # WrappedWireVector (base class) will forward all attribute and method
            # accesses on this _WireStruct to the concatenated WireVector.
            if (
                class_name in kwargs
                and isinstance(kwargs[class_name], int)
                and concatenated_type is WireVector
            ):
                # Special case: simplify the concatenated type to Const.
                concatenated = Const(
                    bitwidth=self._bitwidth,
                    name=name,
                    block=block,
                    val=kwargs[class_name],
                )
            else:
                concatenated = concatenated_type(
                    bitwidth=self._bitwidth, name=name, block=block
                )
            super().__init__(wire=concatenated)

            # self._components maps from component name to each component's WireVector.
            components = {}
            self.__dict__["_components"] = components

            # Handle Input and Register special cases.
            if concatenated_type is Input or concatenated_type is Register:
                kwargs = {class_name: None}
            elif component_type is Input or component_type is Register:
                kwargs = {component_meta.name: None for component_meta in schema}

            if class_name in kwargs:
                # Check for unused kwargs.
                for component_name in kwargs:
                    if component_name != class_name:
                        msg = (
                            "Do not pass additional kwargs to @wire_struct when "
                            f'slicing. ("{class_name}" was passed so don\'t pass '
                            f'"{component_name}")'
                        )
                        raise PyrtlError(msg)
                # Concatenated value was provided. Slice it into components.
                _slice(
                    block=block,
                    schema=schema,
                    bitwidth=self._bitwidth,
                    component_type=component_type,
                    name=name,
                    concatenated=concatenated,
                    components=components,
                    concatenated_value=kwargs[class_name],
                )
            else:
                # Component values were provided; concatenate them.
                # Check that values were provided for all components.
                expected_component_names = [
                    component_meta.name for component_meta in schema
                ]
                for expected_component_name in expected_component_names:
                    if expected_component_name not in kwargs:
                        msg = (
                            "You must provide kwargs for all @wire_struct components "
                            "when concatenating (missing kwarg "
                            f'"{expected_component_name}")'
                        )
                        raise PyrtlError(msg)
                # Check for unused kwargs.
                for component_name in kwargs:
                    if component_name not in expected_component_names:
                        msg = (
                            "Do not pass additional kwargs to @wire_struct when "
                            f'concatenating (don\'t pass "{component_name}")'
                        )
                        raise PyrtlError(msg)

                _concatenate(
                    block=block,
                    schema=schema,
                    component_type=component_type,
                    name=name,
                    concatenated=concatenated,
                    components=components,
                    component_map=kwargs,
                )

        def __getattr__(self, component_name: str):
            """Retrieve a component by name.

            Components are concatenated to form the concatenated :class:`WireVector`, or
            sliced from the concatenated :class:`WireVector`.

            :param component_name: The name of the component wire.
            """
            components = self.__dict__["_components"]
            if component_name in components:
                return components[component_name]
            return super().__getattr__(component_name)

        def __len__(self):
            components = self.__dict__["_components"]
            return len(components)

    return _WireStruct


def wire_matrix(component_schema, size: int):
    """Returns a class that assigns numbered indices to :class:`WireVector` slices.

    ``wire_matrix`` assigns numbered indices to *non-overlapping* :class:`WireVector`
    slices. ``wire_matrix`` is very similar to :func:`wire_struct`, so read
    :func:`wire_struct`'s documentation first.

    .. note::

        See `example1.2-wire-struct.py
        <https://github.com/UCSBarchlab/PyRTL/blob/development/examples/example1.2-wire-struct.py>`_
        for additional examples.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    An example 32-bit ``Word`` ``wire_matrix``, which represents a group of four bytes,
    can be defined as::

        >>> Word = wire_matrix(component_schema=8, size=4)

    .. NOTE::

        ``wire_matrix`` returns a class, like :func:`~collections.namedtuple`.

    Construction
    ------------

    Once a ``wire_matrix`` class is defined, it can be instantiated by providing drivers
    for all of its wires. This can be done in two ways::

        # Provide a driver for each component, most significant bits first.
        >>> word = Word(values=[0x89, 0xAB, 0xCD, 0xEF])

        # Provide a driver for all components.
        >>> word = Word(values=[0x89ABCDEF])

    .. NOTE::

        When specifying drivers for each component, the most significant bits are
        specified first.

    After instantiating a ``wire_matrix``, regardless of how it was constructed, the
    instance functions as a :class:`WireVector` containing all the wires, so ``word``
    functions as a :class:`WireVector` with bitwidth 32. The named slice can be accessed
    with square brackets (``__getitem__``), for example ``word[0]`` and ``word[3]``,
    which both function as :class:`WireVector` with bitwidth 8. ``word[0]`` refers to
    the most significant byte, and ``word[3]`` refers to the least significant byte.
    Both the instance and the slices are first-class :class:`WireVector`, so they can be
    manipulated with all the usual PyRTL operators.

    Naming
    ------

    A ``wire_matrix`` can be assigned a name in the usual way::

        # The whole Word is named 'w', so the components will have names
        # w[0], w[1], ...
        >>> word = Word(name="w", values=[0x89, 0xAB, 0xCD, 0xEF])
        >>> word[0].name
        'w[0]'

    Composition
    -----------

    .. doctest comment::

        These `Composition` examples are currently not ``doctested`` because the
        ``doctest`` environment converts all class attributes to strings, due to its use
        of ``exec``. In the example below, the attribute will be "Byte" (``str``),
        instead of ``Byte`` (``class``), which breaks ``wire_struct``.

    ``wire_matrix`` can be composed with itself and :func:`wire_struct`. For example, we
    can define some multi-dimensional byte arrays::

        Array1D = wire_matrix(component_schema=8, size=2)
        Array2D = wire_matrix(component_schema=Array1D, size=2)

    Drivers must be specified for all components, but they can be specified at any
    level. All these examples construct an equivalent ``wire_matrix``::

        array_2d = Array2D(values=[0x89AB, 0xCDEF])
        array_2d = Array2D(values=[Array1D(values=[0x89, 0xAB]),
                                   0xCDEF])
        array_2d = Array2D(values=[Array1D(values=[0x89, 0xAB]),
                                   Array1D(values=[0xCD, 0xEF])])

    Accessing Slices
    ----------------

    Hierarchical components are accessed by composing ``[]`` operators, for example::

        print(array_2d[0][0].bitwidth)  # Prints 8.
        print(array_2d[0][1].bitwidth)  # Prints 8.

    When ``wire_matrix`` is composed with :func:`wire_struct`, components can be
    accessed by combining the ``[]`` and ``.`` operators::

        @wire_struct
        class Byte:
            high: 4
            low: 4
        Array1D = wire_matrix(component_schema=Byte, size=2)
        array_1d = Array1D(values=[0xAB, 0xCD])

        print(array_1d[0].high.bitwidth)  # Prints 4.

    ``len()`` returns the number of components in the ``wire_matrix``::

        print(len(array_1d))  # Prints '2'.

    Types
    -----

    You can change the type of a ``wire_matrix``'s components with the
    ``component_type`` constructor argument::

        # Generates Outputs named ``output_word[0]``, ``output_word[1]``, ...
        >>> word = Word(name="output_word",
        ...             component_type=pyrtl.Output,
        ...             values=[0x89ABCDEF])

        >>> isinstance(word[1], pyrtl.Output)
        True
        >>> word[1].name
        'output_word[1]'

    You can change the type of the ``wire_matrix`` itself with the ``concatenated_type``
    cnstructor argument::

        # Generates an Input named ``input_word``.
        >>> word = Word(name="input_word", concatenated_type=pyrtl.Input)

    .. NOTE::

        No values are specified for ``input_word`` because its value is not known until
        simulation time.
    """
    # Determine each component's bitwidth.
    if hasattr(component_schema, "_is_wire_struct") or hasattr(
        component_schema, "_is_wire_matrix"
    ):
        component_bitwidth = component_schema._bitwidth
    else:
        component_bitwidth = component_schema
        component_schema = None

    class _WireMatrix(WrappedWireVector):
        _component_bitwidth = component_bitwidth
        _component_schema = component_schema
        _size = size

        _bitwidth = component_bitwidth * size
        _is_wire_matrix = True

        def __init__(
            self,
            name: str = "",
            block: Block = None,
            concatenated_type=WireVector,
            component_type=WireVector,
            values: list | None = None,
        ):
            # The concatenated WireVector contains all the _WireMatrix's wires.
            # WrappedWireVector (base class) will forward all attribute and method
            # accesses on this _WireMatrix to the concatenated WireVector.
            if values is None:
                values = []
            if (
                len(values) == 1
                and isinstance(values[0], int)
                and concatenated_type is WireVector
            ):
                # Special case: simplify the concatenated type to Const.
                concatenated = Const(
                    bitwidth=self._bitwidth, name=name, block=block, val=values[0]
                )
            else:
                concatenated = concatenated_type(
                    bitwidth=self._bitwidth, name=name, block=block
                )
            super().__init__(wire=concatenated)

            schema = []
            for component_name in range(self._size):
                schema.append(
                    _ComponentMeta(
                        name=component_name,
                        bitwidth=self._component_bitwidth,
                        type=component_schema,
                    )
                )

            # By default, slice the concatenated value into components iff exactly one
            # value is provided.
            slicing = len(values) == 1

            # Handle Input and Register special cases.
            if concatenated_type is Input or concatenated_type is Register:
                # Slice the concatenated value. Override the default 'slicing' because
                # 'values' is empty when slicing a concatenated Input or Register.
                #
                # Note that we can't just check len(values) == 1 after we set values to
                # [None] because that doesn't work when there is only one element in the
                # wire_matrix. We must distinguish between:
                #
                # 1. Slicing values to produce values[0] (this case).
                # 2. Concatenating values[0] to produce values (next case).
                #
                # But len(values) == 1 in both cases. The slice in (1) and concatenate
                # in (2) are both no-ops, but we have to get the direction right. In the
                # first case, values[0] is driven by values, and in the second case,
                # values is driven by values[0].
                slicing = True
                values = [None]
            elif component_type is Input or component_type is Register:
                values = [None for _ in range(self._size)]

            self._components = [None for i in range(len(schema))]
            if slicing:
                # Concatenated value was provided. Slice it into components.
                _slice(
                    block=block,
                    schema=schema,
                    bitwidth=self._bitwidth,
                    component_type=component_type,
                    name=name,
                    concatenated=concatenated,
                    components=self._components,
                    concatenated_value=values[0],
                )
            else:
                if len(values) != len(schema):
                    msg = (
                        "wire_matrix constructor expects 1 value to slice, or "
                        f"{len(schema)} values to concatenate (received {len(values)} "
                        "values)"
                    )
                    raise PyrtlError(msg)
                # Component values were provided; concatenate them.
                _concatenate(
                    block=block,
                    schema=schema,
                    component_type=component_type,
                    name=name,
                    concatenated=concatenated,
                    components=self._components,
                    component_map=values,
                )

        def __getitem__(self, key):
            return self._components[key]

        def __len__(self):
            return len(self._components)

    return _WireMatrix


def one_hot_to_binary(w: WireVectorLike) -> WireVector:
    """Takes a one-hot input and returns the bit position of the high bit in binary.

    If the input contains multiple ``1``'s, the smallest bit position containing a ``1``
    will be returned. If the input contains no ``1``'s, ``0`` will be returned.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Examples::

        >>> input = pyrtl.Input(name="input", bitwidth=8)
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.one_hot_to_binary(input)
        >>> output.bitwidth
        3
        >>> 2 ** 3 == input.bitwidth
        True

        >>> sim = pyrtl.Simulation()
        >>> sim.step(provided_inputs={"input": 0b0010})
        >>> sim.inspect("output")
        1

        >>> sim.step(provided_inputs={"input": 64})
        >>> sim.inspect("output")
        6
        >>> 2 ** 6
        64

        >>> # Bit positions 2 and 3 contain 1's, but 2 is smaller.
        >>> sim.step(provided_inputs={"input": 0b1100})
        >>> sim.inspect("output")
        2

        >>> sim.step(provided_inputs={"input": 0})
        >>> sim.inspect("output")
        0

    :param w: A :class:`WireVector`, or any type that can be coerced to
              :class:`WireVector` by :func:`as_wires`.

    :return: a :class:`WireVector` containing the smallest bit position of the high bit,
             in binary.
    """

    w = as_wires(w)

    pos = 0  # Bit position of the first 1
    already_found = as_wires(False)  # True if first 1 already found, False otherwise

    for i in range(len(w)):
        pos = select(w[i] & ~already_found, i, pos)
        already_found = already_found | w[i]

    return pos


def binary_to_one_hot(
    bit_position: WireVectorLike, max_bitwidth: int | None = None
) -> WireVector:
    """Given a ``bit_position``, return a :class:`WireVector` with only that bit set to
    ``1``.

    If the ``max_bitwidth`` provided is not large enough for the given ``bit_position``,
    a ``0``-valued :class:`WireVector` of size ``max_bitwidth`` will be returned.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Examples::

        >>> input = pyrtl.Input(name="input", bitwidth=4)
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.binary_to_one_hot(input)
        >>> output.bitwidth
        16
        >>> 16 == 2 ** input.bitwidth
        True

        >>> sim = pyrtl.Simulation()
        >>> sim.step(provided_inputs={"input": 0})
        >>> bin(sim.inspect("output"))
        '0b1'

        >>> sim.step(provided_inputs={"input": 3})
        >>> bin(sim.inspect("output"))
        '0b1000'

    :param bit_position: A :class:`WireVector`, or any type that can be coerced to
        :class:`WireVector` by :func:`as_wires`.
    :param max_bitwidth: Optional integer maximum bitwidth for the resulting one-hot
        :class:`WireVector`.

    :return: A :class:`WireVector` with the bit at ``bit_position`` set to ``1`` and all
             other bits set to ``0``. Bit position 0 is the least significant bit.
    """

    bit_position = as_wires(bit_position)

    if max_bitwidth is not None:
        bitwidth = max_bitwidth
    else:
        bitwidth = 2 ** len(bit_position)

    # Need to dynamically set the appropriate bit position since bit_position may not be
    # a Const
    return shift_left_logical(Const(1, bitwidth=bitwidth), bit_position)



import doctest
import unittest

import pyrtl


class TestDocTests(unittest.TestCase):
    """Test documentation examples."""

    def test_doctests(self):
        failures, tests = doctest.testmod(m=pyrtl.memory)
        self.assertGreater(tests, 0)
        self.assertEqual(failures, 0)


class RTLMemBlockDesignBase(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()
        self.bitwidth = 3
        self.addrwidth = 5
        self.output1 = pyrtl.Output(self.bitwidth, "output1")
        self.output2 = pyrtl.Output(self.bitwidth, "output2")
        self.mem_read_address1 = pyrtl.Input(self.addrwidth, name="mem_read_address1")
        self.mem_read_address2 = pyrtl.Input(self.addrwidth, name="mem_read_address2")
        self.mem_write_address = pyrtl.Input(self.addrwidth, name="mem_write_address")
        self.mem_write_data = pyrtl.Input(self.bitwidth, name="mem_write_data")
        self.memory = pyrtl.MemBlock(
            bitwidth=self.bitwidth,
            addrwidth=self.addrwidth,
            name="self.memory",
            max_read_ports=None,
        )

    def tearDown(self):
        pyrtl.reset_working_block()

    def test_memblock_simple(self):
        self.output1 <<= self.memory[self.mem_read_address1]
        self.output2 <<= self.memory[self.mem_read_address2]
        self.memory[self.mem_write_address] <<= self.mem_write_data
        pyrtl.working_block().sanity_check()

    def test_memblock_assign_with_extension(self):
        big_output = pyrtl.Output(self.bitwidth + 1, "big_output")
        big_output <<= self.memory[self.mem_read_address1]
        self.output1 <<= 1
        self.output2 <<= 2
        self.memory[self.mem_write_address] <<= self.mem_write_data
        pyrtl.working_block().sanity_check()

    def test_memblock_with_write_enable_with_equalsign(self):
        we = pyrtl.Const(1, bitwidth=1)
        self.output1 <<= self.memory[self.mem_read_address1]
        self.output2 <<= self.memory[self.mem_read_address2]
        self.memory[self.mem_write_address] <<= pyrtl.MemBlock.EnabledWrite(
            self.mem_write_data, enable=we
        )
        pyrtl.working_block().sanity_check()

    def test_memblock_direct_assignment_error(self):
        with self.assertRaises(pyrtl.PyrtlError):
            self.memory[self.mem_write_address] = self.mem_write_data

    def test_memblock_connection_with_ints(self):
        self.memory[self.mem_write_address] <<= 5

    # test does not check functionality, just that it will generate hardware
    def test_memblock_to_memblock_direct_operation(self):
        temp = (
            self.memory[self.mem_read_address1] == self.memory[self.mem_read_address2]
        )
        temp = (
            self.memory[self.mem_read_address1] != self.memory[self.mem_read_address2]
        )
        temp = self.memory[self.mem_read_address1] & self.memory[self.mem_read_address2]
        temp = self.memory[self.mem_read_address1] | self.memory[self.mem_read_address2]
        temp = self.memory[self.mem_read_address1] + self.memory[self.mem_read_address2]
        temp = self.memory[self.mem_read_address1] - self.memory[self.mem_read_address2]
        temp2 = (
            self.memory[self.mem_read_address1] * self.memory[self.mem_read_address2]
        )
        self.output1 <<= temp
        self.output2 <<= temp2
        pyrtl.working_block().sanity_check()

    def test_2read_1write(self):
        small_memory = pyrtl.MemBlock(
            bitwidth=self.bitwidth,
            addrwidth=self.addrwidth,
            name="small_memory",
            max_read_ports=2,
            max_write_ports=1,
        )
        temp = small_memory[self.mem_read_address1]  # read
        temp2 = small_memory[self.mem_read_address2]  # read

        self.output1 <<= temp
        self.output2 <<= temp2
        small_memory[self.mem_write_address] <<= pyrtl.Const(6)  # write
        pyrtl.working_block().sanity_check()

    def test_over_max_read_ports(self):
        lim_memory = pyrtl.MemBlock(
            bitwidth=self.bitwidth,
            addrwidth=self.addrwidth,
            name="lim_memory",
            max_read_ports=8,
        )
        for _i in range(lim_memory.max_read_ports):
            self.output1 <<= lim_memory[self.mem_read_address1]
        with self.assertRaises(pyrtl.PyrtlError):
            self.output2 <<= lim_memory[self.mem_read_address2]

    def test_over_max_write_ports(self):
        lim_memory = pyrtl.MemBlock(
            bitwidth=self.bitwidth,
            addrwidth=self.addrwidth,
            name="lim_memory",
            max_write_ports=4,
        )
        for _i in range(lim_memory.max_write_ports):
            lim_memory[self.mem_write_address] <<= pyrtl.Const(6)
        with self.assertRaises(pyrtl.PyrtlError):
            lim_memory[self.mem_write_address] <<= pyrtl.Const(6)

    def test_memblock_added_user_named(self):
        mem_name = "small_memory"
        small_memory = pyrtl.MemBlock(
            bitwidth=self.bitwidth,
            addrwidth=self.addrwidth,
            name=mem_name,
            max_read_ports=2,
            max_write_ports=1,
        )
        self.assertIs(
            pyrtl.working_block().get_memblock_by_name(mem_name), small_memory
        )

    def test_memblock_added_default_named(self):
        mem = pyrtl.MemBlock(32, 8)
        self.assertIs(pyrtl.working_block().get_memblock_by_name(mem.name), mem)


class RTLMemBlockErrorTests(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_negative_bitwidth(self):
        with self.assertRaises(pyrtl.PyrtlError):
            pyrtl.MemBlock(bitwidth=-1, addrwidth=1)

    def test_negative_addrwidth(self):
        with self.assertRaises(pyrtl.PyrtlError):
            pyrtl.MemBlock(bitwidth=1, addrwidth=-1)

    def test_memindex_bitwidth_greater_than_addrwidth(self):
        mem = pyrtl.MemBlock(bitwidth=1, addrwidth=1)
        mem_addr = pyrtl.Input(2, "mem_addr")
        mem_out = pyrtl.Output(1, "mem_out")
        with self.assertRaises(pyrtl.PyrtlError):
            mem_out <<= mem[mem_addr]

    def test_memblock_write_data_larger_than_memory_bitwidth(self):
        mem = pyrtl.MemBlock(bitwidth=1, addrwidth=1)
        mem_addr = pyrtl.Input(1, "mem_addr")
        mem_in = pyrtl.Input(2, "mem_in")
        with self.assertRaises(pyrtl.PyrtlError):
            mem[mem_addr] <<= mem_in

    def test_memblock_enable_signal_not_1_bit(self):
        mem = pyrtl.MemBlock(bitwidth=1, addrwidth=1)
        mem_addr = pyrtl.Input(1, "mem_addr")
        mem_in = pyrtl.Input(1, "mem_in")
        with self.assertRaises(pyrtl.PyrtlError):
            mem[mem_addr] <<= pyrtl.MemBlock.EnabledWrite(mem_in, enable=pyrtl.Input(2))

    def test_read_ports_exception(self):
        mem = pyrtl.MemBlock(bitwidth=1, addrwidth=1)
        with self.assertRaises(pyrtl.PyrtlError):
            mem.read_ports()


class MemIndexedTests(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_memindexed_name(self):
        self.mem = pyrtl.MemBlock(8, 8)
        x = self.mem[2]
        x.name = "test_name"
        self.assertEqual(x.name, "test_name")
        self.assertEqual(x.wire.name, "test_name")

    def test_read_memindexed_ilshift(self):
        self.mem = pyrtl.MemBlock(8, 8)
        self.mem_val_map = {self.mem: {0: 5, 1: 4, 2: 3, 3: 2, 4: 1, 5: 0}}
        a = pyrtl.Input(3)
        x = self.mem[a]
        y = pyrtl.Output(8, "y")
        z = pyrtl.Output(8, "z")
        y <<= x
        z <<= x
        sim = pyrtl.Simulation(memory_value_map=self.mem_val_map)
        for i in range(5):
            sim.step({a: i})
            self.assertEqual(sim.inspect(y), 5 - i)
            self.assertEqual(sim.inspect(z), 5 - i)
        self.assertEqual(self.mem.num_read_ports, 1)

    def test_write_memindexed_ilshift(self):
        self.mem1 = pyrtl.MemBlock(8, 8)
        self.mem2 = pyrtl.MemBlock(8, 8, asynchronous=True)
        self.mem_val_map = {self.mem1: {0: 0, 1: 1, 2: 2, 3: 3, 4: 4, 5: 5}}
        addr1 = pyrtl.Input(3)
        addr2 = pyrtl.Input(3)  # will be one behind addr1
        inp = pyrtl.Input(3)
        x = self.mem1[addr1]  # value follows addr1
        self.mem2[x] <<= inp
        out = pyrtl.Output(9, name="out")
        out <<= self.mem2[addr2]  # one behind addr1, so one behind x
        sim = pyrtl.Simulation(memory_value_map=self.mem_val_map)
        for i in range(5):
            sim.step(
                {
                    addr1: i,
                    addr2: 0 if i == 0 else i - 1,  # one behind addr1
                    inp: 5 - i,
                }
            )
            self.assertEqual(sim.inspect(out), 0 if i == 0 else 5 - (i - 1))
        self.assertEqual(self.mem1.num_read_ports, 1)  # 2 b/c of the output read
        self.assertEqual(self.mem2.num_write_ports, 1)

    def test_read_memindexed_ior(self):
        self.mem = pyrtl.MemBlock(8, 8)
        self.mem_val_map = {self.mem: {0: 5, 1: 4, 2: 3, 3: 2, 4: 1, 5: 0}}
        decide = pyrtl.Input(1)
        ind = pyrtl.Input(3)
        x = self.mem[ind]
        y = pyrtl.Output(8, "y")
        z = pyrtl.Output(8, "z")
        w = pyrtl.Output(8, "w")
        with pyrtl.conditional_assignment:
            with decide:
                y |= x
                z |= x
            with pyrtl.otherwise:
                w |= x
        sim = pyrtl.Simulation(memory_value_map=self.mem_val_map)
        for i in range(5):
            sim.step({decide: i % 2, ind: i})
            if i == 0:
                y_exp, z_exp, w_exp = 0, 0, 5
            elif i == 1:
                y_exp, z_exp, w_exp = 4, 4, 0
            elif i == 2:
                y_exp, z_exp, w_exp = 0, 0, 3
            elif i == 3:
                y_exp, z_exp, w_exp = 2, 2, 0
            else:
                y_exp, z_exp, w_exp = 0, 0, 1
            self.assertEqual(sim.inspect(y), y_exp)
            self.assertEqual(sim.inspect(z), z_exp)
            self.assertEqual(sim.inspect(w), w_exp)
        self.assertEqual(self.mem.num_read_ports, 1)

    def test_write_memindexed_ior(self):
        self.mem1 = pyrtl.MemBlock(8, 8)
        self.mem2 = pyrtl.MemBlock(8, 8, asynchronous=True)
        self.mem_val_map = {self.mem1: {0: 0, 1: 1, 2: 2, 3: 3, 4: 4, 5: 5}}
        decide = pyrtl.Input(1)
        inp = pyrtl.Input(3)
        addr1 = pyrtl.Input(3)
        addr2 = pyrtl.Input(3)  # will be one behind addr1
        zero = pyrtl.Const(0, 3)
        x = self.mem1[addr1]
        x.name = "x"
        out = pyrtl.Output(8, name="out")
        with pyrtl.conditional_assignment:
            with decide:
                self.mem2[x] |= inp
            with pyrtl.otherwise:
                self.mem2[x] |= zero
        out <<= self.mem2[addr2]  # one behind addr1, so one behind x
        sim = pyrtl.Simulation(memory_value_map=self.mem_val_map)
        for i in range(5):
            sim.step(
                {
                    decide: i % 2,
                    addr1: i,
                    addr2: 0 if i == 0 else i - 1,  # one behind addr1
                    inp: 5 - i,
                }
            )
            if (i == 0) | (i == 1) | (i == 3):
                out_exp = 0
            elif i == 2:
                out_exp = 4
            else:
                out_exp = 2
            self.assertEqual(sim.inspect(out), out_exp)
        self.assertEqual(self.mem1.num_read_ports, 1)
        self.assertEqual(self.mem2.num_write_ports, 1)

    def test_memindexed_len(self):
        self.mem = pyrtl.MemBlock(bitwidth=8, addrwidth=1)
        self.assertEqual(len(self.mem[0]), 8)
        self.mem_2 = pyrtl.MemBlock(bitwidth=16, addrwidth=1)
        self.assertEqual(len(self.mem_2[0]), 16)

    def test_memindexed_getitem(self):
        mem = pyrtl.MemBlock(bitwidth=8, addrwidth=1, max_read_ports=None)
        mem_addr = pyrtl.Input(1, "mem_addr")
        mem_out_array = [pyrtl.Output(8, "mem_out_" + str(i)) for i in range(8)]
        for i in range(8):
            mem_out_array[i] <<= mem[mem_addr][i]
        mem_value_map = {mem: {0: 7, 1: 5}}
        sim = pyrtl.Simulation(memory_value_map=mem_value_map)
        sim.step({mem_addr: 0})
        binary = bin(mem_value_map[mem][0])[2:].zfill(8)
        self.assertEqual(
            [sim.inspect(mem_out_array[j]) for j in range(8)],
            [int(binary[7 - j]) for j in range(8)],
        )

    def test_memindexed_sign_extended(self):
        mem = pyrtl.MemBlock(bitwidth=8, addrwidth=1)
        mem_addr = pyrtl.Input(1, "mem_addr")
        mem_out = pyrtl.Output(16, "mem_out")
        mem_out <<= mem[mem_addr].sign_extended(16)
        mem_value_map = {mem: {0: 0b00101101, 1: 0b10011011}}
        mem_value_map_sign_extended = [0b0000000000101101, 0b1111111110011011]
        sim = pyrtl.Simulation(memory_value_map=mem_value_map)
        sim.step({mem_addr: 0})
        self.assertEqual(sim.inspect(mem_out), mem_value_map_sign_extended[0])
        sim.step({mem_addr: 1})
        self.assertEqual(sim.inspect(mem_out), mem_value_map_sign_extended[1])

    def test_memindexed_zero_extended(self):
        mem = pyrtl.MemBlock(bitwidth=8, addrwidth=1)
        mem_addr = pyrtl.Input(1, "mem_addr")
        mem_out = pyrtl.Output(16, "mem_out")
        mem_out <<= mem[mem_addr].zero_extended(16)
        mem_value_map = {mem: {0: 0b00101101, 1: 0b10011011}}
        mem_value_map_zero_extended = [0b0000000000101101, 0b0000000010011011]
        sim = pyrtl.Simulation(memory_value_map=mem_value_map)
        sim.step({mem_addr: 0})
        self.assertEqual(sim.inspect(mem_out), mem_value_map_zero_extended[0])
        sim.step({mem_addr: 1})
        self.assertEqual(sim.inspect(mem_out), mem_value_map_zero_extended[1])


class RTLRomBlockWiring(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()
        self.bitwidth = 3
        self.addrwidth = 5
        self.output1 = pyrtl.Output(self.bitwidth, "output1")
        self.in1 = pyrtl.Input(self.addrwidth, name="mem_write_address")
        self.in2 = pyrtl.Input(self.addrwidth, name="mem_write_address")
        self.data = list(range(2**5))
        self.memory = pyrtl.RomBlock(
            bitwidth=self.bitwidth,
            addrwidth=self.addrwidth,
            name="self.memory",
            romdata=self.data,
            max_read_ports=None,
        )

    def tearDown(self):
        pyrtl.reset_working_block()

    def test_read(self):
        self.output1 <<= self.memory[self.in1]

    def test_direct_assignment_error(self):
        with self.assertRaises(pyrtl.PyrtlError):
            self.memory[self.in1] = self.in2

    def test_int_index_error(self):
        with self.assertRaises(pyrtl.PyrtlError):
            _ = self.memory[3]

    def test_other_non_wire_index_error(self):
        with self.assertRaises(pyrtl.PyrtlError):
            _ = self.memory[()]
        with self.assertRaises(pyrtl.PyrtlError):
            _ = self.memory["test"]
        with self.assertRaises(pyrtl.PyrtlError):
            _ = self.memory["15"]
        with self.assertRaises(pyrtl.PyrtlError):
            _ = self.memory[False]

    def test_write(self):
        with self.assertRaises(pyrtl.PyrtlError):
            self.memory[self.in1] <<= 5

    # test does not check functionality, just that it will generate hardware
    def test_rom_to_rom_direct_operation(self):
        temp = self.memory[self.in1] == self.memory[self.in2]
        temp = self.memory[self.in1] != self.memory[self.in2]  # != creates two nets
        temp = self.memory[self.in1] & self.memory[self.in2]
        temp = self.memory[self.in1] | self.memory[self.in2]
        temp = self.memory[self.in1] + self.memory[self.in2]
        temp = self.memory[self.in1] - self.memory[self.in2]
        temp = self.memory[self.in1] * self.memory[self.in2]
        block = pyrtl.working_block()
        self.assertEqual(len(block.logic), 22)
        self.output1 <<= temp


class RTLRomGetReadData(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    @staticmethod
    def sample_roms():
        def rom_func(address):
            return (2 * address + 1) % 8

        rom = pyrtl.RomBlock(3, 3, [2, 4, 7, 1])
        romf = pyrtl.RomBlock(3, 3, rom_func)
        return rom, romf

    def invalid_rom_read(self, rom, address):
        with self.assertRaises(pyrtl.PyrtlError):
            rom._get_read_data(address)

    def test_invalid_address(self):
        for rom in self.sample_roms():
            self.invalid_rom_read(rom, -1)
            self.invalid_rom_read(rom, 8)
            self.invalid_rom_read(rom, 5809)

    def test_invalid_address_types(self):
        for rom in self.sample_roms():
            self.invalid_rom_read(rom, "test")
            self.invalid_rom_read(rom, pyrtl.Const(10))
            self.invalid_rom_read(rom, [])
            self.invalid_rom_read(rom, slice(1, 3))
            # self.invalid_rom_read(rom, False)  # should this be valid?

    def test_invalid_value_function(self):
        def bad_func(address):
            return str(address)

        def bad_func_2(address):
            return pyrtl.Const(address)

        rom1 = pyrtl.RomBlock(5, 5, ["test", ()])
        rom2 = pyrtl.RomBlock(5, 5, [pyrtl.Const(0), bad_func])
        romf1 = pyrtl.RomBlock(5, 5, bad_func)
        romf2 = pyrtl.RomBlock(5, 5, bad_func_2)
        for rom in (rom1, rom2, romf1, romf2):
            self.invalid_rom_read(rom, 0)
            self.invalid_rom_read(rom, 1)

    def test_value_out_of_range(self):
        def rom_func(address):
            return 2 * (8 - address) + 1

        rom1 = pyrtl.RomBlock(3, 3, [15, 8, 7, 1])
        romf1 = pyrtl.RomBlock(3, 3, rom_func)

        for rom in (rom1, romf1):
            self.invalid_rom_read(rom, 0)
            self.invalid_rom_read(rom, 1)

    def test_out_of_range(self):
        for rom in self.sample_roms():
            self.invalid_rom_read(rom, -1)
            self.invalid_rom_read(rom, 8)
            self.invalid_rom_read(rom, 5809)

    def test_over_max_read_ports(self):
        width = 6
        rom = pyrtl.RomBlock(width, width, [2, 4, 7, 1])
        for _i in range(rom.max_read_ports):
            rom_read_address = pyrtl.Input(width)
            rom_out = pyrtl.Output(width)
            rom_out <<= rom[rom_read_address]
        rom_read_address = pyrtl.Input(width)
        rom_out = pyrtl.Output(width)
        with self.assertRaises(pyrtl.PyrtlError):
            rom_out <<= rom[rom_read_address]

    def test_valid_get_read(self):
        rom, romf = self.sample_roms()
        for address, expected in enumerate((2, 4, 7, 1)):
            self.assertEqual(rom._get_read_data(address), expected)
        for address, expected in enumerate((1, 3, 5, 7, 1)):
            self.assertEqual(romf._get_read_data(address), expected)

    def test_build_new_roms(self):
        width = 6
        rom = pyrtl.RomBlock(6, 6, [2, 4, 7, 1], build_new_roms=True)
        for _i in range(width):
            rom_read_address = pyrtl.Input(width)
            rom_out = pyrtl.Output(width)
            rom_out <<= rom[rom_read_address]
        roms = set()
        for romNet in pyrtl.working_block().logic_subset("m"):
            curr_rom = romNet.op_param[1]
            roms.add(curr_rom)
        self.assertEqual(len(roms), 3)


class TestSignedRomBlockData(unittest.TestCase):
    def setUp(self):
        pyrtl.reset_working_block()

    def test_negative_romdata(self):
        romdata = [-4, -3, -2, -1, 0, 1, 2, 3]
        rom = pyrtl.RomBlock(addrwidth=3, bitwidth=3, romdata=romdata)
        counter = pyrtl.Register(bitwidth=rom.addrwidth)
        counter.next <<= counter + 1
        read_data = pyrtl.Output(name="read_data", bitwidth=rom.bitwidth)
        read_data <<= rom[counter]
        sim = pyrtl.Simulation()
        for i in range(2**rom.addrwidth):
            sim.step()
            actual_read_data = pyrtl.val_to_signed_integer(
                sim.inspect("read_data"), bitwidth=rom.bitwidth
            )
            self.assertEqual(actual_read_data, romdata[i])

    def test_unsigned_romdata(self):
        romdata = [0, 1, 2, 3, 4, 5, 6, 7]
        rom = pyrtl.RomBlock(addrwidth=3, bitwidth=3, romdata=romdata)
        counter = pyrtl.Register(bitwidth=rom.addrwidth)
        counter.next <<= counter + 1
        read_data = pyrtl.Output(name="read_data", bitwidth=rom.bitwidth)
        read_data <<= rom[counter]
        sim = pyrtl.Simulation()
        for i in range(2**rom.addrwidth):
            sim.step()
            actual_read_data = sim.inspect("read_data")
            self.assertEqual(actual_read_data, romdata[i])


if __name__ == "__main__":
    unittest.main()
